"""Tidy3d package imports"""

from __future__ import annotations

from tidy3d.components.boundary import BroadbandModeABCFitterParam, BroadbandModeABCSpec
from tidy3d.components.data.index import SimulationDataMap
from tidy3d.components.index import SimulationMap
from tidy3d.components.material.multi_physics import MultiPhysicsMedium
from tidy3d.components.material.tcad.charge import (
    ChargeConductorMedium,
    ChargeInsulatorMedium,
    SemiconductorMedium,
)
from tidy3d.components.material.tcad.heat import (
    FluidMedium,
    FluidSpec,
    SolidMedium,
    SolidSpec,
)
from tidy3d.components.microwave.data.monitor_data import (
    AntennaMetricsData,
)
from tidy3d.components.spice.analysis.dc import (
    ChargeToleranceSpec,
    IsothermalSteadyChargeDCAnalysis,
)
from tidy3d.components.spice.sources.dc import DCCurrentSource, DCVoltageSource
from tidy3d.components.spice.sources.types import VoltageSourceType
from tidy3d.components.tcad.analysis.heat_simulation_type import UnsteadyHeatAnalysis, UnsteadySpec
from tidy3d.components.tcad.boundary.heat import VerticalNaturalConvectionCoeffModel
from tidy3d.components.tcad.boundary.specification import (
    HeatBoundarySpec,
    HeatChargeBoundarySpec,
)
from tidy3d.components.tcad.data.monitor_data.mesh import VolumeMeshData
from tidy3d.components.tcad.data.sim_data import (
    DeviceCharacteristics,
    HeatChargeSimulationData,
    HeatSimulationData,
    VolumeMesherData,
)
from tidy3d.components.tcad.data.types import (
    SteadyCapacitanceData,
    SteadyCurrentDensityData,
    SteadyElectricFieldData,
    SteadyEnergyBandData,
    SteadyFreeCarrierData,
    SteadyPotentialData,
    TemperatureData,
)
from tidy3d.components.tcad.doping import ConstantDoping, GaussianDoping
from tidy3d.components.tcad.generation_recombination import FossumCarrierLifetime
from tidy3d.components.tcad.grid import (
    DistanceUnstructuredGrid,
    GridRefinementLine,
    GridRefinementRegion,
    UniformUnstructuredGrid,
)
from tidy3d.components.tcad.mesher import VolumeMesher
from tidy3d.components.tcad.monitors.charge import (
    SteadyCapacitanceMonitor,
    SteadyCurrentDensityMonitor,
    SteadyElectricFieldMonitor,
    SteadyEnergyBandMonitor,
    SteadyFreeCarrierMonitor,
    SteadyPotentialMonitor,
)
from tidy3d.components.tcad.monitors.heat import TemperatureMonitor
from tidy3d.components.tcad.monitors.mesh import VolumeMeshMonitor
from tidy3d.components.tcad.simulation.heat import HeatSimulation
from tidy3d.components.tcad.simulation.heat_charge import HeatChargeSimulation
from tidy3d.components.tcad.types import (
    AugerRecombination,
    CaugheyThomasMobility,
    ConstantMobilityModel,
    ConvectionBC,
    CurrentBC,
    DistributedGeneration,
    HeatFluxBC,
    HeatFromElectricSource,
    HeatSource,
    InsulatingBC,
    RadiativeRecombination,
    ShockleyReedHallRecombination,
    SlotboomBandGapNarrowing,
    TemperatureBC,
    UniformHeatSource,
    VoltageBC,
)

from .components.apodization import ApodizationSpec

# boundary placement for other solvers
# boundary placement for other solvers
from .components.bc_placement import (
    MediumMediumInterface,
    SimulationBoundary,
    StructureBoundary,
    StructureSimulationBoundary,
    StructureStructureInterface,
)

# analytic beams
from .components.beam import (
    AstigmaticGaussianBeamProfile,
    GaussianBeamProfile,
    PlaneWaveBeamProfile,
)

# boundary
from .components.boundary import (
    PML,
    ABCBoundary,
    Absorber,
    AbsorberParams,
    BlochBoundary,
    Boundary,
    BoundaryEdge,
    BoundaryEdgeType,
    BoundarySpec,
    DefaultAbsorberParameters,
    DefaultPMLParameters,
    DefaultStablePMLParameters,
    InternalAbsorber,
    ModeABCBoundary,
    PECBoundary,
    Periodic,
    PMCBoundary,
    PMLParams,
    PMLTypes,
    StablePML,
)

# data
from .components.data.data_array import (
    CellDataArray,
    ChargeDataArray,
    DiffractionDataArray,
    EMECoefficientDataArray,
    EMEModeIndexDataArray,
    EMEScalarFieldDataArray,
    EMEScalarModeFieldDataArray,
    EMESMatrixDataArray,
    FieldProjectionAngleDataArray,
    FieldProjectionCartesianDataArray,
    FieldProjectionKSpaceDataArray,
    FluxDataArray,
    FluxTimeDataArray,
    HeatDataArray,
    IndexedDataArray,
    IndexedFieldVoltageDataArray,
    IndexedTimeDataArray,
    IndexedVoltageDataArray,
    ModeAmpsDataArray,
    ModeIndexDataArray,
    PointDataArray,
    ScalarFieldDataArray,
    ScalarFieldTimeDataArray,
    ScalarModeFieldCylindricalDataArray,
    ScalarModeFieldDataArray,
    SpatialDataArray,
    SpatialVoltageDataArray,
    SteadyVoltageDataArray,
)
from .components.data.dataset import (
    FieldDataset,
    FieldTimeDataset,
    ModeSolverDataset,
    PermittivityDataset,
)
from .components.data.monitor_data import (
    AbstractFieldProjectionData,
    AuxFieldTimeData,
    DiffractionData,
    DirectivityData,
    FieldData,
    FieldProjectionAngleData,
    FieldProjectionCartesianData,
    FieldProjectionKSpaceData,
    FieldTimeData,
    FluxData,
    FluxTimeData,
    ModeData,
    ModeSolverData,
    PermittivityData,
)
from .components.data.sim_data import DATA_TYPE_MAP, SimulationData
from .components.data.utils import (
    TetrahedralGridDataset,
    TriangularGridDataset,
)
from .components.eme.data.dataset import (
    EMECoefficientDataset,
    EMEFieldDataset,
    EMEModeSolverDataset,
    EMESMatrixDataset,
)
from .components.eme.data.monitor_data import EMECoefficientData, EMEFieldData, EMEModeSolverData
from .components.eme.data.sim_data import EMESimulationData
from .components.eme.grid import (
    EMECompositeGrid,
    EMEExplicitGrid,
    EMEGrid,
    EMEModeSpec,
    EMEUniformGrid,
)
from .components.eme.monitor import (
    EMECoefficientMonitor,
    EMEFieldMonitor,
    EMEModeSolverMonitor,
    EMEMonitor,
)

# EME
from .components.eme.simulation import EMESimulation
from .components.eme.sweep import EMEFreqSweep, EMELengthSweep, EMEModeSweep, EMEPeriodicitySweep

# field projection
from .components.field_projection import FieldProjector

# frequency conversion utilities
from .components.frequencies import FreqRange, FrequencyUtils, frequencies, wavelengths

# geometry
from .components.geometry.base import Box, ClipOperation, Geometry, GeometryGroup, Transformed
from .components.geometry.mesh import TriangleMesh
from .components.geometry.polyslab import PolySlab
from .components.geometry.primitives import Cylinder, Sphere
from .components.grid.corner_finder import CornerFinderSpec
from .components.grid.grid import Coords, Coords1D, FieldGrid, Grid, YeeGrid
from .components.grid.grid_spec import (
    AutoGrid,
    CustomGrid,
    CustomGridBoundaries,
    GridRefinement,
    GridSpec,
    LayerRefinementSpec,
    QuasiUniformGrid,
    UniformGrid,
)

# lumped elements
from .components.lumped_element import (
    AdmittanceNetwork,
    CoaxialLumpedResistor,
    LinearLumpedElement,
    LumpedElement,
    LumpedResistor,
    RectangularLumpedElement,
    RLCNetwork,
)

# medium
# for docs
from .components.medium import (
    PEC,
    PEC2D,
    PMC,
    AbstractMedium,
    AnisotropicMedium,
    CustomAnisotropicMedium,
    CustomDebye,
    CustomDrude,
    CustomLorentz,
    CustomMedium,
    CustomPoleResidue,
    CustomSellmeier,
    Debye,
    Drude,
    FullyAnisotropicMedium,
    HammerstadSurfaceRoughness,
    HuraySurfaceRoughness,
    KerrNonlinearity,
    Lorentz,
    LossyMetalMedium,
    Medium,
    Medium2D,
    NonlinearModel,
    NonlinearSpec,
    NonlinearSusceptibility,
    PECMedium,
    PerturbationMedium,
    PerturbationPoleResidue,
    PMCMedium,
    PoleResidue,
    Sellmeier,
    SurfaceImpedanceFitterParam,
    TwoPhotonAbsorption,
    medium_from_nk,
)
from .components.mode.data.sim_data import ModeSimulationData

# Mode
from .components.mode.simulation import ModeSimulation

# modes
from .components.mode_spec import ModeSpec

# monitors
from .components.monitor import (
    AuxFieldTimeMonitor,
    DiffractionMonitor,
    DirectivityMonitor,
    FieldMonitor,
    FieldProjectionAngleMonitor,
    FieldProjectionCartesianMonitor,
    FieldProjectionKSpaceMonitor,
    FieldProjectionSurface,
    FieldTimeMonitor,
    FluxMonitor,
    FluxTimeMonitor,
    ModeMonitor,
    ModeSolverMonitor,
    Monitor,
    PermittivityMonitor,
)
from .components.parameter_perturbation import (
    CustomChargePerturbation,
    CustomHeatPerturbation,
    IndexPerturbation,
    LinearChargePerturbation,
    LinearHeatPerturbation,
    NedeljkovicSorefMashanovich,
    ParameterPerturbation,
    PermittivityPerturbation,
)

# run time spec
from .components.run_time_spec import RunTimeSpec

# scene
# scene
from .components.scene import Scene

# simulation
from .components.simulation import Simulation
from .components.source.base import Source
from .components.source.current import (
    CustomCurrentSource,
    PointDipole,
    UniformCurrentSource,
)
from .components.source.field import (
    TFSF,
    AstigmaticGaussianBeam,
    CustomFieldSource,
    FixedAngleSpec,
    FixedInPlaneKSpec,
    GaussianBeam,
    ModeSource,
    PlaneWave,
)
from .components.source.frame import (
    PECFrame,
)

# sources
from .components.source.time import (
    ContinuousWave,
    CustomSourceTime,
    GaussianPulse,
    SourceTime,
)

# structures
from .components.structure import MeshOverrideStructure, Structure

# subpixel
from .components.subpixel_spec import (
    ContourPathAveraging,
    HeuristicPECStaircasing,
    PECConformal,
    PolarizedAveraging,
    Staircasing,
    SubpixelSpec,
    SurfaceImpedance,
    VolumetricAveraging,
)

# time modulation
from .components.time_modulation import (
    ContinuousWaveTimeModulation,
    ModulationSpec,
    SpaceModulation,
    SpaceTimeModulation,
)
from .components.transformation import RotationAroundAxis
from .components.viz import VisualizationSpec, restore_matplotlib_rcparams

# config
from .config import config

# constants imported as `C_0 = td.C_0` or `td.constants.C_0`
from .constants import C_0, EPSILON_0, ETA_0, HBAR, K_B, MU_0, Q_e, inf
from .log import log, set_logging_console, set_logging_file

# material library dict imported as `from tidy3d import material_library`
# get material `mat` and variant `var` as `material_library[mat][var]`
from .material_library.material_library import material_library
from .material_library.parametric_materials import Graphene

# updater
from .updater import Updater

# version
from .version import __version__


def set_logging_level(level: str) -> None:
    """Raise a warning here instead of setting the logging level."""
    raise DeprecationWarning(
        "``set_logging_level`` no longer supported. "
        f"To set the logging level, call ``tidy3d.config.logging_level = {level}``."
    )


log.info(f"Using client version: {__version__}")

Transformed.update_forward_refs()
ClipOperation.update_forward_refs()
GeometryGroup.update_forward_refs()

__all__ = [
    "C_0",
    "DATA_TYPE_MAP",
    "EPSILON_0",
    "ETA_0",
    "HBAR",
    "K_B",
    "MU_0",
    "PEC",
    "PEC2D",
    "PMC",
    "PML",
    "TFSF",
    "ABCBoundary",
    "Absorber",
    "AbsorberParams",
    "AbstractFieldProjectionData",
    "AbstractMedium",
    "AdmittanceNetwork",
    "AnisotropicMedium",
    "AntennaMetricsData",
    "ApodizationSpec",
    "AstigmaticGaussianBeam",
    "AstigmaticGaussianBeamProfile",
    "AugerRecombination",
    "AutoGrid",
    "AuxFieldTimeData",
    "AuxFieldTimeMonitor",
    "BlochBoundary",
    "Boundary",
    "BoundaryEdge",
    "BoundaryEdgeType",
    "BoundarySpec",
    "Box",
    "BroadbandModeABCFitterParam",
    "BroadbandModeABCSpec",
    "CaugheyThomasMobility",
    "CellDataArray",
    "ChargeConductorMedium",
    "ChargeDataArray",
    "ChargeInsulatorMedium",
    "ChargeToleranceSpec",
    "ClipOperation",
    "CoaxialLumpedResistor",
    "ConstantDoping",
    "ConstantMobilityModel",
    "ContinuousWave",
    "ContinuousWaveTimeModulation",
    "ContourPathAveraging",
    "ConvectionBC",
    "Coords",
    "Coords1D",
    "CornerFinderSpec",
    "CurrentBC",
    "CustomAnisotropicMedium",
    "CustomChargePerturbation",
    "CustomCurrentSource",
    "CustomDebye",
    "CustomDrude",
    "CustomFieldSource",
    "CustomGrid",
    "CustomGridBoundaries",
    "CustomHeatPerturbation",
    "CustomLorentz",
    "CustomMedium",
    "CustomPoleResidue",
    "CustomSellmeier",
    "CustomSourceTime",
    "Cylinder",
    "DCCurrentSource",
    "DCVoltageSource",
    "Debye",
    "DefaultAbsorberParameters",
    "DefaultPMLParameters",
    "DefaultStablePMLParameters",
    "DeviceCharacteristics",
    "DiffractionData",
    "DiffractionDataArray",
    "DiffractionMonitor",
    "DirectivityData",
    "DirectivityMonitor",
    "DistanceUnstructuredGrid",
    "DistributedGeneration",
    "Drude",
    "EMECoefficientData",
    "EMECoefficientDataArray",
    "EMECoefficientDataset",
    "EMECoefficientMonitor",
    "EMECompositeGrid",
    "EMEExplicitGrid",
    "EMEFieldData",
    "EMEFieldDataset",
    "EMEFieldMonitor",
    "EMEFreqSweep",
    "EMEGrid",
    "EMELengthSweep",
    "EMEModeIndexDataArray",
    "EMEModeSolverData",
    "EMEModeSolverDataset",
    "EMEModeSolverMonitor",
    "EMEModeSpec",
    "EMEModeSweep",
    "EMEMonitor",
    "EMEPeriodicitySweep",
    "EMESMatrixDataArray",
    "EMESMatrixDataset",
    "EMEScalarFieldDataArray",
    "EMEScalarModeFieldDataArray",
    "EMESimulation",
    "EMESimulationData",
    "EMESweepSpec",
    "EMEUniformGrid",
    "FieldData",
    "FieldDataset",
    "FieldGrid",
    "FieldMonitor",
    "FieldProjectionAngleData",
    "FieldProjectionAngleDataArray",
    "FieldProjectionAngleMonitor",
    "FieldProjectionCartesianData",
    "FieldProjectionCartesianDataArray",
    "FieldProjectionCartesianMonitor",
    "FieldProjectionKSpaceData",
    "FieldProjectionKSpaceDataArray",
    "FieldProjectionKSpaceMonitor",
    "FieldProjectionSurface",
    "FieldProjector",
    "FieldTimeData",
    "FieldTimeDataset",
    "FieldTimeMonitor",
    "FixedAngleSpec",
    "FixedInPlaneKSpec",
    "FluidMedium",
    "FluidSpec",
    "FluxData",
    "FluxDataArray",
    "FluxMonitor",
    "FluxTimeData",
    "FluxTimeDataArray",
    "FluxTimeMonitor",
    "FossumCarrierLifetime",
    "FreqRange",
    "FrequencyUtils",
    "FullyAnisotropicMedium",
    "GaussianBeam",
    "GaussianBeamProfile",
    "GaussianDoping",
    "GaussianPulse",
    "Geometry",
    "GeometryGroup",
    "Graphene",
    "Grid",
    "GridRefinement",
    "GridRefinementLine",
    "GridRefinementRegion",
    "GridSpec",
    "HammerstadSurfaceRoughness",
    "HeatBoundarySpec",
    "HeatChargeBoundarySpec",
    "HeatChargeSimulation",
    "HeatChargeSimulationData",
    "HeatDataArray",
    "HeatFluxBC",
    "HeatFromElectricSource",
    "HeatSimulation",
    "HeatSimulationData",
    "HeatSource",
    "HeuristicPECStaircasing",
    "HuraySurfaceRoughness",
    "IndexPerturbation",
    "IndexedDataArray",
    "IndexedFieldVoltageDataArray",
    "IndexedTimeDataArray",
    "IndexedVoltageDataArray",
    "InsulatingBC",
    "InternalAbsorber",
    "IsothermalSteadyChargeDCAnalysis",
    "KerrNonlinearity",
    "LayerRefinementSpec",
    "LinearChargePerturbation",
    "LinearHeatPerturbation",
    "LinearLumpedElement",
    "Lorentz",
    "LossyMetalMedium",
    "LumpedElement",
    "LumpedResistor",
    "Medium",
    "Medium2D",
    "MediumMediumInterface",
    "MeshOverrideStructure",
    "ModeABCBoundary",
    "ModeAmpsDataArray",
    "ModeData",
    "ModeIndexDataArray",
    "ModeMonitor",
    "ModeSimulation",
    "ModeSimulationData",
    "ModeSolverData",
    "ModeSolverDataset",
    "ModeSolverMonitor",
    "ModeSource",
    "ModeSpec",
    "ModulationSpec",
    "Monitor",
    "MultiPhysicsMedium",
    "NedeljkovicSorefMashanovich",
    "NonlinearModel",
    "NonlinearSpec",
    "NonlinearSusceptibility",
    "PECBoundary",
    "PECConformal",
    "PECFrame",
    "PECMedium",
    "PMCBoundary",
    "PMCMedium",
    "PMLParams",
    "PMLTypes",
    "ParameterPerturbation",
    "Periodic",
    "PermittivityData",
    "PermittivityDataset",
    "PermittivityMonitor",
    "PermittivityPerturbation",
    "PerturbationMedium",
    "PerturbationPoleResidue",
    "PlaneWave",
    "PlaneWaveBeamProfile",
    "PointDataArray",
    "PointDipole",
    "PolarizedAveraging",
    "PoleResidue",
    "PolySlab",
    "Q_e",
    "QuasiUniformGrid",
    "RLCNetwork",
    "RadiativeRecombination",
    "RectangularLumpedElement",
    "RotationAroundAxis",
    "RunTimeSpec",
    "ScalarFieldDataArray",
    "ScalarFieldTimeDataArray",
    "ScalarModeFieldCylindricalDataArray",
    "ScalarModeFieldDataArray",
    "Scene",
    "Sellmeier",
    "SemiconductorMedium",
    "ShockleyReedHallRecombination",
    "Simulation",
    "SimulationBoundary",
    "SimulationData",
    "SimulationDataMap",
    "SimulationMap",
    "SlotboomBandGapNarrowing",
    "SolidMedium",
    "SolidSpec",
    "Source",
    "SourceTime",
    "SpaceModulation",
    "SpaceTimeModulation",
    "SpatialDataArray",
    "SpatialVoltageDataArray",
    "Sphere",
    "StablePML",
    "Staircasing",
    "SteadyCapacitanceData",
    "SteadyCapacitanceMonitor",
    "SteadyCurrentDensityData",
    "SteadyCurrentDensityMonitor",
    "SteadyElectricFieldData",
    "SteadyElectricFieldMonitor",
    "SteadyEnergyBandData",
    "SteadyEnergyBandMonitor",
    "SteadyFreeCarrierData",
    "SteadyFreeCarrierMonitor",
    "SteadyPotentialData",
    "SteadyPotentialMonitor",
    "SteadyVoltageDataArray",
    "Structure",
    "StructureBoundary",
    "StructureSimulationBoundary",
    "StructureStructureInterface",
    "SubpixelSpec",
    "SurfaceImpedance",
    "SurfaceImpedanceFitterParam",
    "TemperatureBC",
    "TemperatureData",
    "TemperatureMonitor",
    "TetrahedralGridDataset",
    "Transformed",
    "TriangleMesh",
    "TriangularGridDataset",
    "TwoPhotonAbsorption",
    "UniformCurrentSource",
    "UniformGrid",
    "UniformHeatSource",
    "UniformUnstructuredGrid",
    "UnsteadyHeatAnalysis",
    "UnsteadySpec",
    "Updater",
    "VerticalNaturalConvectionCoeffModel",
    "VisualizationSpec",
    "VoltageBC",
    "VoltageSourceType",
    "VolumeMeshData",
    "VolumeMeshMonitor",
    "VolumeMesher",
    "VolumeMesherData",
    "VolumetricAveraging",
    "YeeGrid",
    "__version__",
    "config",
    "frequencies",
    "inf",
    "log",
    "material_library",
    "medium_from_nk",
    "restore_matplotlib_rcparams",
    "set_logging_console",
    "set_logging_file",
    "wavelengths",
]



from __future__ import annotations

from shutil import which
from typing import Union

import tidy3d as td


def check_installation(raise_error: bool = False) -> Union[str, None]:
    """Checks if KLayout is installed and added to the system PATH.
    If it is, this returns the path to the executable. Otherwise, returns ``None``.
    Equivalent to $which("klayout") in the terminal.

    Parameters
    ----------
    raise_error : bool
        Whether to raise an error if KLayout is not found. If ``False``, a warning is shown.

    Returns
    -------
    Union[str, None]
        The path to the KLayout executable. If KLayout is not found, returns ``None``.
    """
    path = which("klayout")
    msg = "KLayout was not found in the system PATH. Please ensure KLayout is installed and added to your system PATH before running KLayout."
    if path is None:
        if raise_error:
            raise RuntimeError(msg)
        else:
            td.log.warning(msg)
    return path



# type information for autograd

# utilities for working with autograd
from __future__ import annotations

import copy
import typing

import pydantic.v1 as pd
from autograd.builtins import dict as dict_ag
from autograd.extend import Box, defvjp, primitive

from tidy3d.components.types import ArrayFloat2D, ArrayLike, Complex, Size1D, _add_schema

# add schema to the Box
_add_schema(Box, title="AutogradBox", field_type_str="autograd.tracer.Box")

# make sure Boxes in tidy3d properly define VJPs for copy operations, for computational graph
_copy = primitive(copy.copy)
_deepcopy = primitive(copy.deepcopy)

defvjp(_copy, lambda ans, x: lambda g: _copy(g))
defvjp(_deepcopy, lambda ans, x, memo: lambda g: _deepcopy(g, memo))

Box.__copy__ = lambda v: _copy(v)
Box.__deepcopy__ = lambda v, memo: _deepcopy(v, memo)

# Types for floats, or collections of floats that can also be autograd tracers
TracedFloat = typing.Union[float, Box]
TracedPositiveFloat = typing.Union[pd.PositiveFloat, Box]
TracedSize1D = typing.Union[Size1D, Box]
TracedSize = typing.Union[tuple[TracedSize1D, TracedSize1D, TracedSize1D], Box]
TracedCoordinate = typing.Union[tuple[TracedFloat, TracedFloat, TracedFloat], Box]
TracedVertices = typing.Union[ArrayFloat2D, Box]

# poles
TracedComplex = typing.Union[Complex, Box]
TracedPoleAndResidue = tuple[TracedComplex, TracedComplex]

# The data type that we pass in and out of the web.run() @autograd.primitive
AutogradTraced = typing.Union[Box, ArrayLike]
PathType = tuple[typing.Union[int, str], ...]
AutogradFieldMap = dict_ag[PathType, AutogradTraced]

InterpolationType = typing.Literal["nearest", "linear"]

__all__ = [
    "AutogradFieldMap",
    "AutogradTraced",
    "TracedCoordinate",
    "TracedFloat",
    "TracedSize",
    "TracedSize1D",
    "TracedVertices",
]



"""Tidy3d webapi types."""

from __future__ import annotations

import os
import pathlib
import tempfile
import time
from datetime import datetime
from typing import Callable, Optional, Union

import pydantic.v1 as pd
from botocore.exceptions import ClientError
from pydantic.v1 import Extra, Field, parse_obj_as

import tidy3d as td
from tidy3d.exceptions import ValidationError
from tidy3d.web.common import REFRESH_TIME

from . import http_util
from .cache import FOLDER_CACHE
from .constants import (
    SIM_ERROR_FILE,
    SIM_FILE_HDF5_GZ,
    SIM_LOG_FILE,
    SIMULATION_DATA_HDF5_GZ,
)
from .core_config import get_logger_console
from .environment import Env
from .exceptions import WebError, WebNotFoundError
from .file_util import read_simulation_from_hdf5
from .http_util import get_version as _get_protocol_version
from .http_util import http
from .s3utils import download_file, download_gz_file, upload_file
from .stub import TaskStub
from .task_info import BatchDetail
from .types import PayType, Queryable, ResourceLifecycle, Submittable, Tidy3DResource


class Folder(Tidy3DResource, Queryable, extra=Extra.allow):
    """Tidy3D Folder."""

    folder_id: str = Field(..., title="Folder id", description="folder id", alias="projectId")
    folder_name: str = Field(
        ..., title="Folder name", description="folder name", alias="projectName"
    )

    @classmethod
    def list(cls, projects_endpoint: str = "tidy3d/projects") -> []:
        """List all folders.

        Returns
        -------
        folders : [Folder]
            List of folders
        """
        resp = http.get(projects_endpoint)
        return (
            parse_obj_as(
                list[Folder],
                resp,
            )
            if resp
            else None
        )

    @classmethod
    def get(
        cls,
        folder_name: str,
        create: bool = False,
        projects_endpoint: str = "tidy3d/projects",
        project_endpoint: str = "tidy3d/project",
    ):
        """Get folder by name.

        Parameters
        ----------
        folder_name : str
            Name of the folder.
        create : str
            If the folder doesn't exist, create it.

        Returns
        -------
        folder : Folder
        """
        folder = FOLDER_CACHE.get(folder_name)
        if not folder:
            resp = http.get(project_endpoint, params={"projectName": folder_name})
            if resp:
                folder = Folder(**resp)
        if create and not folder:
            resp = http.post(projects_endpoint, {"projectName": folder_name})
            if resp:
                folder = Folder(**resp)
        FOLDER_CACHE[folder_name] = folder
        return folder

    @classmethod
    def create(cls, folder_name: str):
        """Create a folder, return existing folder if there is one has the same name.

        Parameters
        ----------
        folder_name : str
            Name of the folder.

        Returns
        -------
        folder : Folder
        """
        return Folder.get(folder_name, True)

    def delete(self, projects_endpoint: str = "tidy3d/projects"):
        """Remove this folder."""

        http.delete(f"{projects_endpoint}/{self.folder_id}")

    def delete_old(self, days_old: int) -> int:
        """Remove folder contents older than ``days_old``."""

        return http.delete(
            f"tidy3d/tasks/{self.folder_id}/tasks",
            params={"daysOld": days_old},
        )

    def list_tasks(self, projects_endpoint: str = "tidy3d/projects") -> list[Tidy3DResource]:
        """List all tasks in this folder.

        Returns
        -------
        tasks : List[:class:`.SimulationTask`]
            List of tasks in this folder
        """
        resp = http.get(f"{projects_endpoint}/{self.folder_id}/tasks")
        return (
            parse_obj_as(
                list[SimulationTask],
                resp,
            )
            if resp
            else None
        )


class SimulationTask(ResourceLifecycle, Submittable, extra=Extra.allow):
    """Interface for managing the running of a :class:`.Simulation` task on server."""

    task_id: Optional[str] = Field(
        ...,
        title="task_id",
        description="Task ID number, set when the task is uploaded, leave as None.",
        alias="taskId",
    )
    folder_id: Optional[str] = Field(
        None,
        title="folder_id",
        description="Folder ID number, set when the task is uploaded, leave as None.",
        alias="folderId",
    )
    status: Optional[str] = Field(title="status", description="Simulation task status.")

    real_flex_unit: float = Field(
        None, title="real FlexCredits", description="Billed FlexCredits.", alias="realCost"
    )

    created_at: Optional[datetime] = Field(
        title="created_at", description="Time at which this task was created.", alias="createdAt"
    )

    task_type: Optional[str] = Field(
        title="task_type", description="The type of task.", alias="taskType"
    )

    folder_name: Optional[str] = Field(
        "default",
        title="Folder Name",
        description="Name of the folder associated with this task.",
        alias="folderName",
    )

    callback_url: str = Field(
        None,
        title="Callback URL",
        description="Http PUT url to receive simulation finish event. "
        "The body content is a json file with fields "
        "``{'id', 'status', 'name', 'workUnit', 'solverVersion'}``.",
    )

    # simulation_type: str = pd.Field(
    #     None,
    #     title="Simulation Type",
    #     description="Type of simulation, used internally only.",
    # )

    # parent_tasks: Tuple[TaskId, ...] = pd.Field(
    #     None,
    #     title="Parent Tasks",
    #     description="List of parent task ids for the simulation, used internally only."
    # )

    @pd.root_validator(pre=True)
    def _error_if_jax_sim(cls, values):
        """Raise error if user tries to submit simulation that's a JaxSimulation."""
        sim = values.get("simulation")
        if sim is None:
            return values
        if "JaxSimulation" in str(type(sim)):
            raise ValueError(
                "'JaxSimulation' not compatible with regular webapi functions. "
                "Either convert it to Simulation with 'jax_sim.to_simulation()[0]' or use "
                "the 'adjoint.run' function to run JaxSimulations."
            )
        return values

    @classmethod
    def create(
        cls,
        task_type: str,
        task_name: str,
        folder_name: str = "default",
        callback_url: Optional[str] = None,
        simulation_type: str = "tidy3d",
        parent_tasks: Optional[list[str]] = None,
        file_type: str = "Gz",
        port_name_list: Optional[list[str]] = None,
        projects_endpoint: str = "tidy3d/projects",
    ) -> SimulationTask:
        """Create a new task on the server.

        Parameters
        ----------
        task_type: :class".TaskType"
            The type of task.
        task_name: str
            The name of the task.
        folder_name: str,
            The name of the folder to store the task. Default is "default".
        callback_url: str
            Http PUT url to receive simulation finish event. The body content is a json file with
            fields ``{'id', 'status', 'name', 'workUnit', 'solverVersion'}``.
        simulation_type : str
            Type of simulation being uploaded.
        parent_tasks : List[str]
            List of related task ids.
        file_type: str
            the simulation file type Json, Hdf5, Gz

        Returns
        -------
        :class:`SimulationTask`
            :class:`SimulationTask` object containing info about status, size,
            credits of task and others.
        """

        # handle backwards compatibility, "tidy3d" is the default simulation_type
        if simulation_type is None:
            simulation_type = "tidy3d"

        folder = Folder.get(folder_name, create=True)
        payload = {
            "taskName": task_name,
            "taskType": task_type,
            "callbackUrl": callback_url,
            "simulationType": simulation_type,
            "parentTasks": parent_tasks,
            "fileType": file_type,
        }
        # Component modeler: include port names if provided
        if port_name_list:
            # Align with backend contract: expect 'portNames' (not 'portNameList')
            payload["portNames"] = port_name_list

        resp = http.post(f"{projects_endpoint}/{folder.folder_id}/tasks", payload)
        # RF group creation may return group-level info without 'taskId'.
        # Use 'groupId' (or 'batchId' as fallback) as the resource id for subsequent uploads.
        if "taskId" not in resp and task_type == "RF":
            # Prefer using 'batchId' as the resource id for uploads (S3 STS expects a task-like id).
            if "batchId" in resp:
                resp["taskId"] = resp["batchId"]
            elif "groupId" in resp:
                resp["taskId"] = resp["groupId"]
            else:
                raise WebError("Missing resource ID for task creation. Contact customer support.")
        return SimulationTask(**resp, taskType=task_type, folder_name=folder_name)

    @classmethod
    def get(cls, task_id: str, verbose: bool = True) -> SimulationTask:
        """Get task from the server by id.

        Parameters
        ----------
        task_id: str
            Unique identifier of task on server.
        verbose:
            If `True`, will print progressbars and status, otherwise, will run silently.

        Returns
        -------
        :class:`.SimulationTask`
            :class:`.SimulationTask` object containing info about status,
             size, credits of task and others.
        """
        try:
            resp = http.get(f"tidy3d/tasks/{task_id}/detail")
        except WebNotFoundError as e:
            td.log.error(f"The requested task ID '{task_id}' does not exist.")
            raise e

        task = SimulationTask(**resp) if resp else None
        return task

    @classmethod
    def get_running_tasks(cls) -> list[SimulationTask]:
        """Get a list of running tasks from the server"

        Returns
        -------
        List[:class:`.SimulationTask`]
            :class:`.SimulationTask` object containing info about status,
             size, credits of task and others.
        """
        resp = http.get("tidy3d/py/tasks")
        if not resp:
            return []
        return parse_obj_as(list[SimulationTask], resp)

    def delete(self, versions: bool = False):
        """Delete current task from server.

        Parameters
        ----------
        versions : bool = False
            If ``True``, delete all versions of the task in the task group. Otherwise, delete only the version associated with the current task ID.
        """
        if not self.task_id:
            raise ValueError("Task id not found.")

        task_details = http.get(f"tidy3d/tasks/{self.task_id}")

        if task_details and "groupId" in task_details and "version" in task_details:
            group_id = task_details["groupId"]
            version = task_details["version"]
            if versions:
                http.delete("tidy3d/group", json={"groupIds": [group_id]})
            else:
                http.delete(f"tidy3d/group/{group_id}/versions", json={"versions": [version]})
        else:  # Fallback to old method if we can't get the groupId and version
            http.delete(f"tidy3d/tasks/{self.task_id}")

    def get_simulation_json(self, to_file: str, verbose: bool = True) -> pathlib.Path:
        """Get json file for a :class:`.Simulation` from server.

        Parameters
        ----------
        to_file: str
            Save file to path.
        verbose: bool = True
            Whether to display progress bars.

        Returns
        -------
        path: pathlib.Path
            Path to saved file.
        """
        if not self.task_id:
            raise WebError("Expected field 'task_id' is unset.")

        hdf5_file, hdf5_file_path = tempfile.mkstemp(".hdf5")
        os.close(hdf5_file)
        try:
            self.get_simulation_hdf5(hdf5_file_path)
            if os.path.exists(hdf5_file_path):
                json_string = read_simulation_from_hdf5(hdf5_file_path)
                with open(to_file, "w") as file:
                    # Write the string to the file
                    file.write(json_string.decode("utf-8"))
                    if verbose:
                        console = get_logger_console()
                        console.log(f"Generate {to_file} successfully.")
            else:
                raise WebError("Failed to download simulation.json.")
        finally:
            os.unlink(hdf5_file_path)

    def upload_simulation(
        self,
        stub: TaskStub,
        verbose: bool = True,
        progress_callback: Optional[Callable[[float], None]] = None,
        remote_sim_file: str = SIM_FILE_HDF5_GZ,
    ) -> None:
        """Upload :class:`.Simulation` object to Server.

        Parameters
        ----------
        stub: :class:`TaskStub`
            An instance of TaskStub.
        verbose: bool = True
            Whether to display progress bars.
        progress_callback : Callable[[float], None] = None
            Optional callback function called while uploading the data.
        """
        if not self.task_id:
            raise WebError("Expected field 'task_id' is unset.")
        if not stub:
            raise WebError("Expected field 'simulation' is unset.")
        # Also upload hdf5.gz containing all data.
        file, file_name = tempfile.mkstemp()
        os.close(file)
        try:
            # upload simulation
            # compress .hdf5 to .hdf5.gz
            stub.to_hdf5_gz(file_name)
            upload_file(
                self.task_id,
                file_name,
                remote_sim_file,
                verbose=verbose,
                progress_callback=progress_callback,
            )
        finally:
            os.unlink(file_name)

    def upload_file(
        self,
        local_file: str,
        remote_filename: str,
        verbose: bool = True,
        progress_callback: Optional[Callable[[float], None]] = None,
    ) -> None:
        """
        Upload file to platform. Using this method when the json file is too large to parse
         as :class".simulation".
        Parameters
        ----------
        local_file: str
            local file path.
        remote_filename: str
            file name on the server
        verbose: bool = True
            Whether to display progress bars.
        progress_callback : Callable[[float], None] = None
            Optional callback function called while uploading the data.
        """
        if not self.task_id:
            raise WebError("Expected field 'task_id' is unset.")

        upload_file(
            self.task_id,
            local_file,
            remote_filename,
            verbose=verbose,
            progress_callback=progress_callback,
        )

    def submit(
        self,
        solver_version: Optional[str] = None,
        worker_group: Optional[str] = None,
        pay_type: Union[PayType, str] = PayType.AUTO,
        priority: Optional[int] = None,
    ):
        """Kick off this task.

        It will be uploaded to server before
        starting the task. Otherwise, this method assumes that the Simulation has been uploaded by
        the upload_file function, so the task will be kicked off directly.

        Parameters
        ----------
        solver_version: str = None
            target solver version.
        worker_group: str = None
            worker group
        pay_type: Union[PayType, str] = PayType.AUTO
            Which method to pay the simulation.
        priority: int = None
            Priority of the simulation in the Virtual GPU (vGPU) queue (1 = lowest, 10 = highest).
            It affects only simulations from vGPU licenses and does not impact simulations using FlexCredits.
        """
        pay_type = PayType(pay_type) if not isinstance(pay_type, PayType) else pay_type

        if solver_version:
            protocol_version = None
        else:
            protocol_version = http_util.get_version()

        http.post(
            f"tidy3d/tasks/{self.task_id}/submit",
            {
                "solverVersion": solver_version,
                "workerGroup": worker_group,
                "protocolVersion": protocol_version,
                "enableCaching": Env.current.enable_caching,
                "payType": pay_type.value,
                "priority": priority,
            },
        )

    def estimate_cost(self, solver_version=None) -> float:
        """Compute the maximum flex unit charge for a given task, assuming the simulation runs for
        the full ``run_time``. If early shut-off is triggered, the cost is adjusted proportionately.

        Parameters
        ----------
        solver_version: str
            target solver version.

        Returns
        -------
        flex_unit_cost: float
            estimated cost in FlexCredits
        """
        if not self.task_id:
            raise WebError("Expected field 'task_id' is unset.")

        if solver_version:
            protocol_version = None
        else:
            protocol_version = http_util.get_version()

        resp = http.post(
            f"tidy3d/tasks/{self.task_id}/metadata",
            {
                "solverVersion": solver_version,
                "protocolVersion": protocol_version,
            },
        )
        return resp

    def get_sim_data_hdf5(
        self,
        to_file: str,
        verbose: bool = True,
        progress_callback: Optional[Callable[[float], None]] = None,
        remote_data_file: str = SIMULATION_DATA_HDF5_GZ,
    ) -> pathlib.Path:
        """Get simulation data file from Server.

        Parameters
        ----------
        to_file: str
            Save file to path.
        verbose: bool = True
            Whether to display progress bars.
        progress_callback : Callable[[float], None] = None
            Optional callback function called while downloading the data.

        Returns
        -------
        path: pathlib.Path
            Path to saved file.
        """
        if not self.task_id:
            raise WebError("Expected field 'task_id' is unset.")

        file = None
        try:
            file = download_gz_file(
                resource_id=self.task_id,
                remote_filename=remote_data_file,
                to_file=to_file,
                verbose=verbose,
                progress_callback=progress_callback,
            )
        except ClientError:
            if verbose:
                console = get_logger_console()
                console.log(f"Unable to download '{remote_data_file}'.")

        if not file:
            try:
                file = download_file(
                    resource_id=self.task_id,
                    remote_filename=remote_data_file[:-3],
                    to_file=to_file,
                    verbose=verbose,
                    progress_callback=progress_callback,
                )
            except Exception as e:
                raise WebError(
                    "Failed to download the simulation data file from the server. "
                    "Please confirm that the task was successfully run."
                ) from e

        return file

    def get_simulation_hdf5(
        self,
        to_file: str,
        verbose: bool = True,
        progress_callback: Optional[Callable[[float], None]] = None,
        remote_sim_file: str = SIM_FILE_HDF5_GZ,
    ) -> pathlib.Path:
        """Get simulation.hdf5 file from Server.

        Parameters
        ----------
        to_file: str
            Save file to path.
        verbose: bool = True
            Whether to display progress bars.
        progress_callback : Callable[[float], None] = None
            Optional callback function called while downloading the data.

        Returns
        -------
        path: pathlib.Path
            Path to saved file.
        """
        if not self.task_id:
            raise WebError("Expected field 'task_id' is unset.")

        return download_gz_file(
            resource_id=self.task_id,
            remote_filename=remote_sim_file,
            to_file=to_file,
            verbose=verbose,
            progress_callback=progress_callback,
        )

    def get_running_info(self) -> tuple[float, float]:
        """Gets the % done and field_decay for a running task.

        Returns
        -------
        perc_done : float
            Percentage of run done (in terms of max number of time steps).
            Is ``None`` if run info not available.
        field_decay : float
            Average field intensity normalized to max value (1.0).
            Is ``None`` if run info not available.
        """
        if not self.task_id:
            raise WebError("Expected field 'task_id' is unset.")

        resp = http.get(f"tidy3d/tasks/{self.task_id}/progress")
        perc_done = resp.get("perc_done")
        field_decay = resp.get("field_decay")
        return perc_done, field_decay

    def get_log(
        self,
        to_file: str,
        verbose: bool = True,
        progress_callback: Optional[Callable[[float], None]] = None,
    ) -> pathlib.Path:
        """Get log file from Server.

        Parameters
        ----------
        to_file: str
            Save file to path.
        verbose: bool = True
            Whether to display progress bars.
        progress_callback : Callable[[float], None] = None
            Optional callback function called while downloading the data.

        Returns
        -------
        path: pathlib.Path
            Path to saved file.
        """

        if not self.task_id:
            raise WebError("Expected field 'task_id' is unset.")

        return download_file(
            self.task_id,
            SIM_LOG_FILE,
            to_file=to_file,
            verbose=verbose,
            progress_callback=progress_callback,
        )

    def get_error_json(self, to_file: str, verbose: bool = True) -> pathlib.Path:
        """Get error json file for a :class:`.Simulation` from server.

        Parameters
        ----------
        to_file: str
            Save file to path.
        verbose: bool = True
            Whether to display progress bars.

        Returns
        -------
        path: pathlib.Path
            Path to saved file.
        """
        if not self.task_id:
            raise WebError("Expected field 'task_id' is unset.")

        return download_file(
            self.task_id,
            SIM_ERROR_FILE,
            to_file=to_file,
            verbose=verbose,
        )

    def abort(self):
        """Aborting current task from server."""
        if not self.task_id:
            raise ValueError("Task id not found.")
        return http.put(
            "tidy3d/tasks/abort", json={"taskType": self.task_type, "taskId": self.task_id}
        )

    def validate_post_upload(self, parent_tasks: Optional[list[str]] = None):
        """Perform checks after task is uploaded and metadata is processed."""
        if self.task_type == "HEAT_CHARGE" and parent_tasks:
            try:
                if len(parent_tasks) > 1:
                    raise ValueError(
                        "A single parent 'task_id' corresponding to the task in which the meshing "
                        "was run must be provided."
                    )
                try:
                    # get mesh task info
                    mesh_task = SimulationTask.get(parent_tasks[0], verbose=False)
                    assert mesh_task.task_type == "VOLUME_MESH"
                    assert mesh_task.status == "success"
                    # get up-to-date task info
                    task = SimulationTask.get(self.task_id, verbose=False)
                    if task.fileMd5 != mesh_task.childFileMd5:
                        raise ValidationError(
                            "Simulation stored in parent task 'VolumeMesher' does not match the "
                            "current simulation."
                        )
                except Exception as e:
                    raise ValidationError(
                        "The parent task must be a 'VolumeMesher' task which has been successfully "
                        "run and is associated to the same 'HeatChargeSimulation' as provided here."
                    ) from e

            except Exception as e:
                raise WebError(f"Provided 'parent_tasks' failed validation: {e!s}") from e


class BatchTask:
    """Provides a client-side interface for managing a remote batch task.

    This class acts as a wrapper around the API endpoints for a specific batch,
    allowing users to check, submit, monitor, and download data from it.

    Note:
        The 'batch_type' (e.g., "RF_SWEEP") must be provided by the caller to
        most methods, as it dictates which backend service handles the request.
    """

    def __init__(self, batch_id: str):
        self.batch_id = batch_id

    @staticmethod
    def is_batch(resource_id: str, batch_type: str) -> bool:
        """Checks if a given resource ID corresponds to a valid batch task.

        This is a utility function to verify a batch task's existence before
        instantiating the class.

        Parameters
        ----------
        resource_id : str
            The unique identifier for the resource.
        batch_type : str
            The type of the batch to check (e.g., "RF_SWEEP").

        Returns
        -------
        bool
            ``True`` if the resource is a valid batch task, ``False`` otherwise.
        """
        try:
            # TODO PROPERLY FIXME
            # Disable non critical logs due to check for resourceId, until we have a dedicated API for this
            resp = http.get(
                f"tidy3d/tasks/{resource_id}/batch-detail",
                params={"batchType": batch_type},
                suppress_404=True,
            )
            status = bool(resp and isinstance(resp, dict) and "status" in resp)
            return status
        except Exception:
            return False

    def detail(self, batch_type: str) -> BatchDetail:
        """Fetches the detailed information and status of the batch.

        Parameters
        ----------
        batch_type : str
            The type of the batch (e.g., "RF_SWEEP").

        Returns
        -------
        BatchDetail
            An object containing the batch's latest data.
        """
        resp = http.get(
            f"tidy3d/tasks/{self.batch_id}/batch-detail",
            params={"batchType": batch_type},
        )
        # Some backends may return null for collection fields; coerce to sensible defaults
        if isinstance(resp, dict):
            if resp.get("tasks") is None:
                resp["tasks"] = []
        return BatchDetail(**(resp or {}))

    def check(
        self,
        solver_version: Optional[str] = None,
        protocol_version: Optional[str] = None,
        batch_type: str = "",
    ):
        """Submits a request to validate the batch configuration on the server.

        Parameters
        ----------
        solver_version : Optional[str], default=None
            The version of the solver to use for validation.
        protocol_version : Optional[str], default=None
            The data protocol version. Defaults to the current version.
        batch_type : str, default=""
            The type of the batch (e.g., "RF_SWEEP").

        Returns
        -------
        Any
            The server's response to the check request.
        """
        if protocol_version is None:
            protocol_version = _get_protocol_version()
        return http.post(
            f"tidy3d/projects/{self.batch_id}/batch-check",
            {
                "batchType": batch_type,
                "solverVersion": solver_version,
                "protocolVersion": protocol_version,
            },
        )

    def submit(
        self,
        solver_version: Optional[str] = None,
        protocol_version: Optional[str] = None,
        worker_group: Optional[str] = None,
        batch_type: str = "",
    ):
        """Submits the batch for execution on the server.

        Parameters
        ----------
        solver_version : Optional[str], default=None
            The version of the solver to use for execution.
        protocol_version : Optional[str], default=None
            The data protocol version. Defaults to the current version.
        worker_group : Optional[str], default=None
            Optional identifier for a specific worker group to run on.
        batch_type : str, default=""
            The type of the batch (e.g., "RF_SWEEP").

        Returns
        -------
        Any
            The server's response to the submit request.
        """
        if protocol_version is None:
            protocol_version = _get_protocol_version()
        return http.post(
            f"tidy3d/projects/{self.batch_id}/batch-submit",
            {
                "batchType": batch_type,
                "solverVersion": solver_version,
                "protocolVersion": protocol_version,
                "workerGroup": worker_group,
            },
        )

    def postprocess(
        self,
        solver_version: Optional[str] = None,
        protocol_version: Optional[str] = None,
        worker_group: Optional[str] = None,
        batch_type: str = "",
    ):
        """Initiates post-processing for a completed batch run.

        Parameters
        ----------
        solver_version : Optional[str], default=None
            The version of the solver to use for post-processing.
        protocol_version : Optional[str], default=None
            The data protocol version. Defaults to the current version.
        worker_group : Optional[str], default=None
            Optional identifier for a specific worker group to run on.
        batch_type : str, default=""
            The type of the batch (e.g., "RF_SWEEP").

        Returns
        -------
        Any
            The server's response to the post-process request.
        """
        if protocol_version is None:
            protocol_version = _get_protocol_version()
        return http.post(
            f"tidy3d/projects/{self.batch_id}/postprocess",
            {
                "batchType": batch_type,
                "solverVersion": solver_version,
                "protocolVersion": protocol_version,
                "workerGroup": worker_group,
            },
        )

    def wait_for_validate(
        self, timeout: Optional[float] = None, batch_type: str = ""
    ) -> BatchDetail:
        """Waits for the batch to complete the validation stage by polling its status.

        Parameters
        ----------
        timeout : Optional[float], default=None
            Maximum time in seconds to wait. If ``None``, waits indefinitely.
        batch_type : str, default=""
            The type of the batch (e.g., "RF_SWEEP").

        Returns
        -------
        BatchDetail
            The final object after validation completes or a timeout occurs.

        Notes
        -----
        This method blocks until the batch status is 'validate_success',
        'validate_warn', 'validate_fail', or another terminal state like 'blocked'
        or 'aborted', or until the timeout is reached.
        """
        start = datetime.now().timestamp()
        while True:
            d = self.detail(batch_type=batch_type)
            status = d.totalStatus
            if status in ("validate_success", "validate_warn", "validate_fail"):
                return d
            if status in ("blocked", "aborting", "aborted"):
                return d
            if timeout is not None and (datetime.now().timestamp() - start) > timeout:
                return d
            time.sleep(REFRESH_TIME)

    def wait_for_run(self, timeout: Optional[float] = None, batch_type: str = "") -> BatchDetail:
        """Waits for the batch to complete the execution stage by polling its status.

        Parameters
        ----------
        timeout : Optional[float], default=None
            Maximum time in seconds to wait. If ``None``, waits indefinitely.
        batch_type : str, default=""
            The type of the batch (e.g., "RF_SWEEP").

        Returns
        -------
        BatchDetail
            The final object after the run completes or a timeout occurs.

        Notes
        -----
        This method blocks until the batch status reaches a terminal run state like
        'run_success', 'run_failed', 'diverged', 'blocked', or 'aborted',
        or until the timeout is reached.
        """
        start = datetime.now().timestamp()
        while True:
            d = self.detail(batch_type=batch_type)
            status = d.totalStatus
            if status in (
                "run_success",
                "run_failed",
                "diverged",
                "blocked",
                "aborting",
                "aborted",
            ):
                return d
            if timeout is not None and (datetime.now().timestamp() - start) > timeout:
                return d
            time.sleep(REFRESH_TIME)

    def get_data_hdf5(
        self,
        remote_data_file_gz: str,
        to_file: str,
        verbose: bool = True,
        progress_callback: Optional[Callable[[float], None]] = None,
    ) -> pathlib.Path:
        """Downloads a batch data artifact, with a fallback mechanism.

        Parameters
        ----------
        remote_data_file_gz : str
            Remote gzipped filename to download (e.g., 'output/cm_data.hdf5.gz').
        to_file : str
            Local path where the downloaded file will be saved.
        verbose : bool, default=True
            If ``True``, shows progress logs and messages.
        progress_callback : Optional[Callable[[float], None]], default=None
            Optional callback function for progress updates, which receives the
            download percentage as a float.

        Returns
        -------
        pathlib.Path
            An object pointing to the downloaded local file.

        Raises
        ------
        WebError
            If both the gzipped and uncompressed file downloads fail.

        Notes
        -----
        This method first attempts to download the gzipped version of a file.
        If that fails, it falls back to downloading the uncompressed version.
        """
        file = None
        try:
            file = download_gz_file(
                resource_id=self.batch_id,
                remote_filename=remote_data_file_gz,
                to_file=to_file,
                verbose=verbose,
                progress_callback=progress_callback,
            )
        except ClientError:
            if verbose:
                console = get_logger_console()
                console.log(f"Unable to download '{remote_data_file_gz}'.")

        if not file:
            try:
                file = download_file(
                    resource_id=self.batch_id,
                    remote_filename=remote_data_file_gz[:-3],
                    to_file=to_file,
                    verbose=verbose,
                    progress_callback=progress_callback,
                )
            except Exception as e:
                raise WebError(
                    "Failed to download the batch data file from the server. "
                    "Please confirm that the batch has been successfully postprocessed."
                ) from e

        return file



"""Defines heat-charge material specifications for 'HeatChargeSimulation'"""

from __future__ import annotations

from typing import Union

import pydantic.v1 as pd

from tidy3d.components.data.data_array import SpatialDataArray
from tidy3d.components.tcad.source.abstract import StructureBasedHeatChargeSource
from tidy3d.constants import VOLUMETRIC_HEAT_RATE
from tidy3d.log import log


class HeatSource(StructureBasedHeatChargeSource):
    """Adds a volumetric heat source (heat sink if negative values
    are provided) to specific structures in the scene.

    Example
    -------
    >>> heat_source = HeatSource(rate=1, structures=["box"])
    """

    rate: Union[float, SpatialDataArray] = pd.Field(
        title="Volumetric Heat Rate",
        description="Volumetric rate of heating or cooling (if negative) in units of "
        f"{VOLUMETRIC_HEAT_RATE}.",
        units=VOLUMETRIC_HEAT_RATE,
    )


class UniformHeatSource(HeatSource):
    """Volumetric heat source. This class is deprecated. You can use
    'HeatSource' instead.

    Example
    -------
    >>> heat_source = UniformHeatSource(rate=1, structures=["box"]) # doctest: +SKIP
    """

    # NOTE: wrapper for backwards compatibility.

    @pd.root_validator(skip_on_failure=True)
    def issue_warning_deprecated(cls, values):
        """Issue warning for 'UniformHeatSource'."""
        log.warning(
            "'UniformHeatSource' is deprecated and will be discontinued. You can use "
            "'HeatSource' instead."
        )
        return values



"""Provides lowest level, user-facing interface to server."""

from __future__ import annotations

import json
import os
import tempfile
import time
from typing import Callable, Literal, Optional, Union

from requests import HTTPError
from rich.progress import BarColumn, Progress, TaskProgressColumn, TextColumn, TimeElapsedColumn

from tidy3d.components.medium import AbstractCustomMedium
from tidy3d.components.mode.mode_solver import ModeSolver
from tidy3d.components.mode.simulation import ModeSimulation
from tidy3d.components.types.workflow import WorkflowDataType, WorkflowType
from tidy3d.exceptions import WebError
from tidy3d.log import get_logging_console, log
from tidy3d.plugins.smatrix.component_modelers.terminal import TerminalComponentModeler
from tidy3d.web.core.account import Account
from tidy3d.web.core.constants import (
    CM_DATA_HDF5_GZ,
    MODE_DATA_HDF5_GZ,
    MODE_FILE_HDF5_GZ,
    MODELER_FILE_HDF5_GZ,
    SIM_FILE_HDF5,
    SIM_FILE_HDF5_GZ,
    SIMULATION_DATA_HDF5_GZ,
    TaskId,
)
from tidy3d.web.core.environment import Env
from tidy3d.web.core.exceptions import WebNotFoundError
from tidy3d.web.core.http_util import get_version as _get_protocol_version
from tidy3d.web.core.http_util import http
from tidy3d.web.core.task_core import BatchTask, Folder, SimulationTask
from tidy3d.web.core.task_info import ChargeType, TaskInfo
from tidy3d.web.core.types import PayType

from .connect_util import REFRESH_TIME, get_grid_points_str, get_time_steps_str, wait_for_connection
from .tidy3d_stub import Tidy3dStub, Tidy3dStubData

# time between checking run status
RUN_REFRESH_TIME = 1.0

# file names when uploading to S3
SIM_FILE_JSON = "simulation.json"

# not all solvers are supported yet in GUI
GUI_SUPPORTED_TASK_TYPES = ["FDTD", "MODE_SOLVER", "HEAT", "RF"]

# if a solver is in beta stage, cost is subject to change
BETA_TASK_TYPES = ["HEAT", "EME", "HEAT_CHARGE", "VOLUME_MESH"]

# map task_type to solver name for display
SOLVER_NAME = {
    "FDTD": "FDTD",
    "MODE_SOLVER": "Mode",
    "MODE": "Mode",
    "EME": "EME",
    "HEAT": "Heat",
    "HEAT_CHARGE": "HeatCharge",
    "VOLUME_MESH": "VolumeMesher",
}


def _get_url(task_id: str) -> str:
    """Get the URL for a task on our server."""
    return f"{Env.current.website_endpoint}/workbench?taskId={task_id}"


def _get_folder_url(folder_id: str) -> str:
    """Get the URL for a task folder on our server."""
    return f"{Env.current.website_endpoint}/folders/{folder_id}"


def _get_url_rf(resource_id: str) -> str:
    """Get the RF GUI URL for a modeler/batch group."""
    return f"{Env.current.website_endpoint}/rf?taskId={resource_id}"


def _is_modeler_batch(resource_id: str) -> bool:
    """Detect whether the given id corresponds to a modeler batch resource."""
    return BatchTask.is_batch(resource_id, batch_type="RF_SWEEP")


def _batch_detail(resource_id: str):
    return BatchTask(resource_id).detail(batch_type="RF_SWEEP")


def _task_dict_to_url_bullet_list(data_dict: dict) -> str:
    """
    Converts a dictionary into a string formatted as a bullet point list.

    Args:
      data_dict: The dictionary to convert.

    Returns:
      A string with each key-url/value pair as a bullet point.
    """
    # Use a list comprehension to format each key-value pair
    # and then join them together with newline characters.
    return "\n".join([f"- {key}: '{value}'" for key, value in data_dict.items()])


@wait_for_connection
def run(
    simulation: WorkflowType,
    task_name: str,
    folder_name: str = "default",
    path: str = "simulation_data.hdf5",
    callback_url: Optional[str] = None,
    verbose: bool = True,
    progress_callback_upload: Optional[Callable[[float], None]] = None,
    progress_callback_download: Optional[Callable[[float], None]] = None,
    solver_version: Optional[str] = None,
    worker_group: Optional[str] = None,
    simulation_type: str = "tidy3d",
    parent_tasks: Optional[list[str]] = None,
    reduce_simulation: Literal["auto", True, False] = "auto",
    pay_type: Union[PayType, str] = PayType.AUTO,
    priority: Optional[int] = None,
) -> WorkflowDataType:
    """
    Submits a :class:`.Simulation` to server, starts running, monitors progress, downloads,
    and loads results as a :class:`.WorkflowDataType` object.

    Parameters
    ----------
    simulation : Union[:class:`.Simulation`, :class:`.HeatSimulation`, :class:`.EMESimulation`]
        Simulation to upload to server.
    task_name : str
        Name of task.
    folder_name : str = "default"
        Name of folder to store task on web UI.
    path : str = "simulation_data.hdf5"
        Path to download results file (.hdf5), including filename.
    callback_url : str = None
        Http PUT url to receive simulation finish event. The body content is a json file with
        fields ``{'id', 'status', 'name', 'workUnit', 'solverVersion'}``.
    verbose : bool = True
        If ``True``, will print progressbars and status, otherwise, will run silently.
    simulation_type : str = "tidy3d"
        Type of simulation being uploaded.
    progress_callback_upload : Callable[[float], None] = None
        Optional callback function called when uploading file with ``bytes_in_chunk`` as argument.
    progress_callback_download : Callable[[float], None] = None
        Optional callback function called when downloading file with ``bytes_in_chunk`` as argument.
    solver_version: str = None
        target solver version.
    worker_group: str = None
        worker group
    reduce_simulation : Literal["auto", True, False] = "auto"
        Whether to reduce structures in the simulation to the simulation domain only. Note: currently only implemented for the mode solver.
    pay_type: Union[PayType, str] = PayType.AUTO
        Which method to pay the simulation.
    priority: int = None
        Priority of the simulation in the Virtual GPU (vGPU) queue (1 = lowest, 10 = highest).
        It affects only simulations from vGPU licenses and does not impact simulations using FlexCredits.
    Returns
    -------
    Union[:class:`.SimulationData`, :class:`.HeatSimulationData`, :class:`.EMESimulationData`]
        Object containing solver results for the supplied simulation.

    Notes
    -----

        Submitting a simulation to our cloud server is very easily done by a simple web API call.

        .. code-block:: python

            sim_data = tidy3d.web.api.webapi.run(simulation, task_name='my_task', path='out/data.hdf5')

        The :meth:`tidy3d.web.api.webapi.run()` method shows the simulation progress by default.  When uploading a
        simulation to the server without running it, you can use the :meth:`tidy3d.web.api.webapi.monitor`,
        :meth:`tidy3d.web.api.container.Job.monitor`, or :meth:`tidy3d.web.api.container.Batch.monitor` methods to
        display the progress of your simulation(s).

    Examples
    --------

        To access the original :class:`.Simulation` object that created the simulation data you can use:

        .. code-block:: python

            # Run the simulation.
            sim_data = web.run(simulation, task_name='task_name', path='out/sim.hdf5')

            # Get a copy of the original simulation object.
            sim_copy = sim_data.simulation

    See Also
    --------

    :meth:`tidy3d.web.api.webapi.monitor`
        Print the real time task progress until completion.

    :meth:`tidy3d.web.api.container.Job.monitor`
        Monitor progress of running :class:`Job`.

    :meth:`tidy3d.web.api.container.Batch.monitor`
        Monitor progress of each of the running tasks.
    """
    task_id = upload(
        simulation=simulation,
        task_name=task_name,
        folder_name=folder_name,
        callback_url=callback_url,
        verbose=verbose,
        progress_callback=progress_callback_upload,
        simulation_type=simulation_type,
        parent_tasks=parent_tasks,
        solver_version=solver_version,
        reduce_simulation=reduce_simulation,
    )
    start(
        task_id,
        solver_version=solver_version,
        worker_group=worker_group,
        pay_type=pay_type,
        priority=priority,
    )
    monitor(task_id, verbose=verbose)
    data = load(
        task_id=task_id, path=path, verbose=verbose, progress_callback=progress_callback_download
    )
    if isinstance(simulation, ModeSolver):
        simulation._patch_data(data=data)
    return data


@wait_for_connection
def upload(
    simulation: WorkflowType,
    task_name: str,
    folder_name: str = "default",
    callback_url: Optional[str] = None,
    verbose: bool = True,
    progress_callback: Optional[Callable[[float], None]] = None,
    simulation_type: str = "tidy3d",
    parent_tasks: Optional[list[str]] = None,
    source_required: bool = True,
    solver_version: Optional[str] = None,
    reduce_simulation: Literal["auto", True, False] = "auto",
) -> TaskId:
    """
    Upload simulation to server, but do not start running :class:`.Simulation`.

    Parameters
    ----------
    simulation : Union[:class:`.Simulation`, :class:`.HeatSimulation`, :class:`.EMESimulation`]
        Simulation to upload to server.
    task_name : str
        Name of task.
    folder_name : str
        Name of folder to store task on web UI
    callback_url : str = None
        Http PUT url to receive simulation finish event. The body content is a json file with
        fields ``{'id', 'status', 'name', 'workUnit', 'solverVersion'}``.
    verbose : bool = True
        If ``True``, will print progressbars and status, otherwise, will run silently.
    progress_callback : Callable[[float], None] = None
        Optional callback function called when uploading file with ``bytes_in_chunk`` as argument.
    simulation_type : str = "tidy3d"
        Type of simulation being uploaded.
    parent_tasks : List[str]
        List of related task ids.
    source_required: bool = True
        If ``True``, simulations without sources will raise an error before being uploaded.
    solver_version: str = None
        target solver version.
    reduce_simulation: Literal["auto", True, False] = "auto"
        Whether to reduce structures in the simulation to the simulation domain only. Note: currently only implemented for the mode solver.

    Returns
    -------
    str
        Unique identifier of task on server.


    Notes
    -----

        Once you've created a ``job`` object using :class:`tidy3d.web.api.container.Job`, you can upload it to our servers with:

        .. code-block:: python

            web.upload(simulation, task_name="task_name", verbose=verbose)

        It will not run until you explicitly tell it to do so with :meth:`tidy3d.web.api.webapi.start`.

    """
    console = get_logging_console() if verbose else None

    if isinstance(simulation, (ModeSolver, ModeSimulation)):
        simulation = get_reduced_simulation(simulation, reduce_simulation)

    stub = Tidy3dStub(simulation=simulation)
    stub.validate_pre_upload(source_required=source_required)
    log.debug("Creating task.")

    task_type = stub.get_type()
    # Component modeler compatibility: map to RF task type
    port_name_list = None
    if task_type in ("COMPONENT_MODELER", "TERMINAL_COMPONENT_MODELER"):
        task_type = "RF"
        port_name_list = tuple(simulation.sim_dict.keys())

    task = SimulationTask.create(
        task_type,
        task_name,
        folder_name,
        callback_url,
        simulation_type,
        parent_tasks,
        "Gz",
        port_name_list=port_name_list,
    )

    if task_type == "RF":
        # Prefer the group id if present in the creation response; avoid extra GET.
        group_id = getattr(task, "groupId", None) or getattr(task, "group_id", None)
        if not group_id:
            try:
                detail_task = SimulationTask.get(task.task_id, verbose=False)
                group_id = getattr(detail_task, "groupId", None) or getattr(
                    detail_task, "group_id", None
                )
            except Exception:
                group_id = None
        # Prefer returning batch/group id for downstream batch endpoints
        batch_id = getattr(task, "batchId", None) or getattr(task, "batch_id", None)
        resource_id = batch_id or task.task_id
    else:
        group_id = None
        resource_id = task.task_id

    if verbose:
        console.log(
            f"Created task '{task_name}' with resource_id '{resource_id}' and task_type '{task_type}'."
        )
        if task_type in BETA_TASK_TYPES:
            solver_name = SOLVER_NAME[task_type]
            console.log(
                f"Tidy3D's {solver_name} solver is currently in the beta stage. "
                f"Cost of {solver_name} simulations is subject to change in the future."
            )
        if task_type in GUI_SUPPORTED_TASK_TYPES:
            if (task_type == "RF") and (isinstance(simulation, TerminalComponentModeler)):
                url = _get_url_rf(group_id or resource_id)
                folder_url = _get_folder_url(task.folder_id)
                console.log(f"View task using web UI at [link={url}]'{url}'[/link].")
                console.log(f"Task folder: [link={folder_url}]'{task.folder_name}'[/link].")
            else:
                url = _get_url(resource_id)
                folder_url = _get_folder_url(task.folder_id)
                console.log(f"View task using web UI at [link={url}]'{url}'[/link].")
                console.log(f"Task folder: [link={folder_url}]'{task.folder_name}'[/link].")

    remote_sim_file = SIM_FILE_HDF5_GZ
    if task_type == "MODE_SOLVER":
        remote_sim_file = MODE_FILE_HDF5_GZ
    elif task_type == "RF":
        remote_sim_file = MODELER_FILE_HDF5_GZ

    task.upload_simulation(
        stub=stub,
        verbose=verbose,
        progress_callback=progress_callback,
        remote_sim_file=remote_sim_file,
    )

    if task_type == "RF":
        split_path = "tidy3d/projects/component-modeler-split"
        payload = {
            "batchType": "RF_SWEEP",
            "batchId": resource_id,
            "fileName": "modeler.hdf5.gz",
            "protocolVersion": _get_protocol_version(),
        }
        resp = http.post(split_path, payload)
        if verbose:
            console = get_logging_console()
            console.log(
                f"Child simulation subtasks are being uploaded to \n{_task_dict_to_url_bullet_list(resp)}"
            )
        # split (modeler-specific)
        batch = BatchTask(resource_id)
        # Kick off server-side validation for the RF batch.
        batch.check(solver_version=solver_version, batch_type="RF_SWEEP")
        if verbose:
            # Validation phase
            console.log("Validating component modeler and subtask simulations...")

    estimate_cost(task_id=resource_id, solver_version=solver_version, verbose=verbose)

    task.validate_post_upload(parent_tasks=parent_tasks)

    # log the url for the task in the web UI
    log.debug(f"{Env.current.website_endpoint}/folders/{task.folder_id}/tasks/{resource_id}")
    return resource_id


def get_reduced_simulation(simulation, reduce_simulation):
    """
    Adjust the given simulation object based on the reduce_simulation parameter. Currently only
    implemented for the mode solver.

    Parameters
    ----------
    simulation : Simulation
        The simulation object to be potentially reduced.
    reduce_simulation : Literal["auto", True, False]
        Determines whether to reduce the simulation. If "auto", the function will decide based on
        the presence of custom mediums in the simulation.

    Returns
    -------
    Simulation
        The potentially reduced simulation object.
    """

    """
    TODO: This only works for the mode solver, which is also why `simulation.simulation.scene` is
    used below. After refactor to use the new ModeSimulation, it should be possible to put the call
    to this function outside of the MODE_SOLVER check in the upload function. We could implement
    dummy `reduced_simulation_copy` methods for the other solvers or also implement reductions
    there. Note that if we do the latter we may want to also modify the warning below to only
    happen if there are custom media *and* they extend beyond the simulation domain.
    """
    if reduce_simulation == "auto":
        if isinstance(simulation, ModeSimulation):
            sim_mediums = simulation.scene.mediums
        else:
            sim_mediums = simulation.simulation.scene.mediums
        contains_custom = any(isinstance(med, AbstractCustomMedium) for med in sim_mediums)
        reduce_simulation = contains_custom

        if reduce_simulation:
            log.warning(
                f"The {type(simulation)} object contains custom mediums. It will be "
                "automatically restricted to the solver domain to reduce data for uploading. "
                "To force uploading the original object use 'reduce_simulation=False'."
                " Setting 'reduce_simulation=True' will force simulation reduction in all cases and"
                " silence this warning."
            )
    if reduce_simulation:
        return simulation.reduced_simulation_copy
    return simulation


@wait_for_connection
def get_info(task_id: TaskId, verbose: bool = True) -> TaskInfo:
    """Return information about a task.

    Parameters
    ----------
    task_id : str
        Unique identifier of task on server.  Returned by :meth:`upload`.
    verbose : bool = True
        If ``True``, will print progressbars and status, otherwise, will run silently.
    Returns
    -------
    :class:`TaskInfo`
        Object containing information about status, size, credits of task.
    """
    task = SimulationTask.get(task_id, verbose)
    if not task:
        raise ValueError("Task not found.")
    return TaskInfo(**{"taskId": task.task_id, "taskType": task.task_type, **task.dict()})


@wait_for_connection
def start(
    task_id: TaskId,
    solver_version: Optional[str] = None,
    worker_group: Optional[str] = None,
    pay_type: Union[PayType, str] = PayType.AUTO,
    priority: Optional[int] = None,
) -> None:
    """Start running the simulation associated with task.

    Parameters
    ----------

    task_id : str
        Unique identifier of task on server.  Returned by :meth:`upload`.
    solver_version: str = None
        target solver version.
    worker_group: str = None
        worker group
    pay_type: Union[PayType, str] = PayType.AUTO
        Which method to pay the simulation
    priority: int = None
        Priority of the simulation in the Virtual GPU (vGPU) queue (1 = lowest, 10 = highest).
        It affects only simulations from vGPU licenses and does not impact simulations using FlexCredits.
    Note
    ----
    To monitor progress, can call :meth:`monitor` after starting simulation.
    """

    console = get_logging_console()

    # Component modeler batch path: hide split/check/submit
    if _is_modeler_batch(task_id):
        # split (modeler-specific)
        batch = BatchTask(task_id)
        detail = batch.wait_for_validate(batch_type="RF_SWEEP")
        status = detail.totalStatus
        status_str = status.value
        if status_str in ("validate_success", "validate_warn"):
            console.log("Component modeler batch validation has been successful.")
        if status_str not in ("validate_success", "validate_warn"):
            raise WebError(f"Batch task {task_id} is blocked: {status_str}")
        # Submit batch to start runs after validation
        batch.submit(
            solver_version=solver_version, batch_type="RF_SWEEP", worker_group=worker_group
        )
        return

    if priority is not None and (priority < 1 or priority > 10):
        raise ValueError("Priority must be between '1' and '10' if specified.")
    task = SimulationTask.get(task_id)
    if not task:
        raise ValueError("Task not found.")
    task.submit(
        solver_version=solver_version,
        worker_group=worker_group,
        pay_type=pay_type,
        priority=priority,
    )


@wait_for_connection
def get_run_info(task_id: TaskId) -> tuple[Optional[float], Optional[float]]:
    """Gets the % done and field_decay for a running task.

    Parameters
    ----------
    task_id : str
        Unique identifier of task on server.  Returned by :meth:`upload`.

    Returns
    -------
    perc_done : float
        Percentage of run done (in terms of max number of time steps).
        Is ``None`` if run info not available.
    field_decay : float
        Average field intensity normalized to max value (1.0).
        Is ``None`` if run info not available.
    """
    task = SimulationTask(taskId=task_id)
    return task.get_running_info()


def get_status(task_id) -> str:
    """Get the status of a task. Raises an error if status is "error".

    Parameters
    ----------
    task_id : str
        Unique identifier of task on server.  Returned by :meth:`upload`.
    """
    task_info = get_info(task_id)
    status = task_info.status
    if status == "visualize":
        return "success"
    if status == "error":
        try:
            # Try to obtain the error message
            task = SimulationTask(taskId=task_id)
            with tempfile.NamedTemporaryFile(suffix=".json") as tmp_file:
                task.get_error_json(to_file=tmp_file.name)
                with open(tmp_file.name) as f:
                    error_content = json.load(f)
                    error_msg = error_content["msg"]
        except Exception:
            # If the error message could not be obtained, raise a generic error message
            error_msg = "Error message could not be obtained, please contact customer support."

        raise WebError(f"Error running task {task_id}! {error_msg}")
    return status


def monitor(task_id: TaskId, verbose: bool = True, worker_group: Optional[str] = None) -> None:
    """
    Print the real time task progress until completion.

    Notes
    -----

        To monitor the simulation's progress and wait for its completion, use:

        .. code-block:: python

            tidy3d.web.api.webapi.monitor(job.task_id, verbose=verbose).

    Parameters
    ----------
    task_id : str
        Unique identifier of task on server.  Returned by :meth:`upload`.
    verbose : bool = True
        If ``True``, will print progressbars and status, otherwise, will run silently.

    Note
    ----
    To load results when finished, may call :meth:`load`.
    """

    # Batch/modeler monitoring path
    if _is_modeler_batch(task_id):
        _monitor_modeler_batch(task_id, verbose=verbose, worker_group=worker_group)
        return

    console = get_logging_console() if verbose else None

    task_info = get_info(task_id)

    task_name = task_info.taskName

    task_type = task_info.taskType

    break_statuses = ("success", "error", "diverged", "deleted", "draft", "abort", "aborted")

    def get_estimated_cost() -> float:
        """Get estimated cost, if None, is not ready."""
        task_info = get_info(task_id)
        block_info = task_info.taskBlockInfo
        if block_info and block_info.chargeType == ChargeType.FREE:
            est_flex_unit = 0
            grid_points = block_info.maxGridPoints
            time_steps = block_info.maxTimeSteps
            grid_points_str = get_grid_points_str(grid_points)
            time_steps_str = get_time_steps_str(time_steps)
            console.log(
                f"You are running this simulation for FREE. Your current plan allows"
                f" up to {block_info.maxFreeCount} free non-concurrent simulations per"
                f" day (under {grid_points_str} grid points and {time_steps_str}"
                f" time steps)"
            )
        else:
            est_flex_unit = task_info.estFlexUnit
        return est_flex_unit

    def monitor_preprocess() -> None:
        """Periodically check the status."""
        status = get_status(task_id)
        while status not in break_statuses and status != "running":
            new_status = get_status(task_id)
            if new_status != status:
                status = new_status
                if verbose and status != "running":
                    console.log(f"status = {status}")
            time.sleep(REFRESH_TIME)

    status = get_status(task_id)

    if verbose:
        console.log(f"status = {status}")

    # already done
    if status in break_statuses:
        return

    # preprocessing
    if verbose:
        console.log(
            "To cancel the simulation, use 'web.abort(task_id)' or 'web.delete(task_id)' "
            "or abort/delete the task in the web "
            "UI. Terminating the Python script will not stop the job running on the cloud."
        )
        with console.status(f"[bold green]Waiting for '{task_name}'...", spinner="runner"):
            monitor_preprocess()
    else:
        monitor_preprocess()

    # if the estimated cost is ready, print it
    if verbose:
        get_estimated_cost()
        console.log("starting up solver")

    # while running but before the percentage done is available, keep waiting
    while get_run_info(task_id)[0] is None and get_status(task_id) == "running":
        time.sleep(REFRESH_TIME)

    # while running but percentage done is available
    if verbose:
        # verbose case, update progressbar
        console.log("running solver")
        if task_type == "FDTD":
            with Progress(console=console) as progress:
                pbar_pd = progress.add_task("% done", total=100)
                perc_done, _ = get_run_info(task_id)

                while (
                    perc_done is not None and perc_done < 100 and get_status(task_id) == "running"
                ):
                    perc_done, field_decay = get_run_info(task_id)
                    new_description = f"solver progress (field decay = {field_decay:.2e})"
                    progress.update(pbar_pd, completed=perc_done, description=new_description)
                    time.sleep(RUN_REFRESH_TIME)

                perc_done, field_decay = get_run_info(task_id)
                if perc_done is not None and perc_done < 100 and field_decay > 0:
                    console.log(f"early shutoff detected at {perc_done:1.0f}%, exiting.")

                new_description = f"solver progress (field decay = {field_decay:.2e})"
                progress.update(pbar_pd, completed=100, refresh=True, description=new_description)
        elif task_type == "EME":
            with Progress(console=console) as progress:
                pbar_pd = progress.add_task("% done", total=100)
                perc_done, _ = get_run_info(task_id)

                while (
                    perc_done is not None and perc_done < 100 and get_status(task_id) == "running"
                ):
                    perc_done, _ = get_run_info(task_id)
                    new_description = "solver progress"
                    progress.update(pbar_pd, completed=perc_done, description=new_description)
                    time.sleep(RUN_REFRESH_TIME)

                perc_done, _ = get_run_info(task_id)
                new_description = "solver progress"
                progress.update(pbar_pd, completed=100, refresh=True, description=new_description)
        else:
            while get_status(task_id) == "running":
                perc_done, _ = get_run_info(task_id)
                time.sleep(RUN_REFRESH_TIME)

    else:
        # non-verbose case, just keep checking until status is not running or perc_done >= 100
        perc_done, _ = get_run_info(task_id)
        while perc_done is not None and perc_done < 100 and get_status(task_id) == "running":
            perc_done, field_decay = get_run_info(task_id)
            time.sleep(RUN_REFRESH_TIME)

    # post processing
    if verbose:
        status = get_status(task_id)
        if status != "running":
            console.log(f"status = {status}")

        with console.status(f"[bold green]Finishing '{task_name}'...", spinner="runner"):
            while status not in break_statuses:
                new_status = get_status(task_id)
                if new_status != status:
                    status = new_status
                    console.log(f"status = {status}")
                time.sleep(REFRESH_TIME)

        if task_type in GUI_SUPPORTED_TASK_TYPES:
            url = _get_url(task_id)
            console.log(f"View simulation result at [blue underline][link={url}]'{url}'[/link].")
    else:
        while get_status(task_id) not in break_statuses:
            time.sleep(REFRESH_TIME)


@wait_for_connection
def abort(task_id: TaskId):
    """Abort server-side data associated with task.

    Parameters
    ----------
    task_id : str
        Unique identifier of task on server.  Returned by :meth:`upload`.

    Returns
    -------
    TaskInfo
        Object containing information about status, size, credits of task.
    """
    console = get_logging_console()
    try:
        task = SimulationTask.get(task_id, verbose=False)
        if task:
            task.abort()
            url = _get_url(task.task_id)
            console.log(
                f"Task is aborting. View task using web UI at [link={url}]'{url}'[/link] to check the result."
            )
            return TaskInfo(**{"taskId": task.task_id, **task.dict()})
    except WebNotFoundError:
        pass  # Task not found, might be a batch task

    is_batch = BatchTask.is_batch(task_id, batch_type="RF_SWEEP")
    if is_batch:
        url = _get_url_rf(task_id)
        console.log(
            f"Batch task abortion is not yet supported, contact customer support."
            f" View task using web UI at [link={url}]'{url}'[/link]."
        )
        return

    console.log("Task ID cannot be found to be aborted.")
    return


@wait_for_connection
def download(
    task_id: TaskId,
    path: str = "simulation_data.hdf5",
    verbose: bool = True,
    progress_callback: Optional[Callable[[float], None]] = None,
) -> None:
    """Download results of task to file.

    Parameters
    ----------
    task_id : str
        Unique identifier of task on server.  Returned by :meth:`upload`.
    path : str = "simulation_data.hdf5"
        Download path to .hdf5 data file (including filename).
    verbose : bool = True
        If ``True``, will print progressbars and status, otherwise, will run silently.
    progress_callback : Callable[[float], None] = None
        Optional callback function called when downloading file with ``bytes_in_chunk`` as argument.

    """
    # Component modeler batch download path
    if _is_modeler_batch(task_id):
        # Use a more descriptive default filename for component modeler downloads.
        # If the caller left the default as 'simulation_data.hdf5', prefer 'cm_data.hdf5'.
        if os.path.basename(path) == "simulation_data.hdf5":
            base_dir = os.path.dirname(path) or "."
            path = os.path.join(base_dir, "cm_data.hdf5")

        def _download_cm() -> bool:
            try:
                BatchTask(task_id).get_data_hdf5(
                    remote_data_file_gz=CM_DATA_HDF5_GZ,
                    to_file=path,
                    verbose=verbose,
                    progress_callback=progress_callback,
                )
                return True
            except Exception:
                return False

        if not _download_cm():
            BatchTask(task_id).postprocess(batch_type="RF_SWEEP")
            # wait for postprocess to finish
            while True:
                resp = BatchTask(task_id).detail(batch_type="RF_SWEEP")
                total = resp.totalTask or 0
                post_succ = resp.postprocessSuccess or 0
                status = resp.totalStatus
                status_str = status.value
                if status_str in {"error", "diverged", "blocked", "aborted", "aborting"}:
                    raise WebError(
                        f"Batch task {task_id} failed during postprocess: {status_str}"
                    ) from None
                if total > 0 and post_succ >= total:
                    break
                time.sleep(REFRESH_TIME)
            if not _download_cm():
                raise WebError("Failed to download 'cm_data' after postprocess completion.")
        return

    # Regular single-task download
    task_info = get_info(task_id)
    task_type = task_info.taskType

    remote_data_file = SIMULATION_DATA_HDF5_GZ
    if task_type == "MODE_SOLVER":
        remote_data_file = MODE_DATA_HDF5_GZ

    task = SimulationTask(taskId=task_id)
    task.get_sim_data_hdf5(
        path,
        verbose=verbose,
        progress_callback=progress_callback,
        remote_data_file=remote_data_file,
    )


@wait_for_connection
def download_json(task_id: TaskId, path: str = SIM_FILE_JSON, verbose: bool = True) -> None:
    """Download the ``.json`` file associated with the :class:`.Simulation` of a given task.

    Parameters
    ----------
    task_id : str
        Unique identifier of task on server.  Returned by :meth:`upload`.
    path : str = "simulation.json"
        Download path to .json file of simulation (including filename).
    verbose : bool = True
        If ``True``, will print progressbars and status, otherwise, will run silently.

    """

    task = SimulationTask(taskId=task_id)
    task.get_simulation_json(path, verbose=verbose)


@wait_for_connection
def delete_old(days_old: int, folder_name: str = "default") -> int:
    """Remove folder contents older than ``days_old``."""
    folder = Folder.get(folder_name, create=True)
    return folder.delete_old(days_old)


@wait_for_connection
def load_simulation(
    task_id: TaskId, path: str = SIM_FILE_JSON, verbose: bool = True
) -> WorkflowType:
    """Download the ``.json`` file of a task and load the associated simulation.

    Parameters
    ----------
    task_id : str
        Unique identifier of task on server.  Returned by :meth:`upload`.
    path : str = "simulation.json"
        Download path to .json file of simulation (including filename).
    verbose : bool = True
        If ``True``, will print progressbars and status, otherwise, will run silently.

    Returns
    -------
    Union[:class:`.Simulation`, :class:`.HeatSimulation`, :class:`.EMESimulation`]
        Simulation loaded from downloaded json file.
    """

    task = SimulationTask.get(task_id)
    task.get_simulation_json(path, verbose=verbose)
    return Tidy3dStub.from_file(path)


@wait_for_connection
def download_log(
    task_id: TaskId,
    path: str = "tidy3d.log",
    verbose: bool = True,
    progress_callback: Optional[Callable[[float], None]] = None,
) -> None:
    """Download the tidy3d log file associated with a task.

    Parameters
    ----------
    task_id : str
        Unique identifier of task on server.  Returned by :meth:`upload`.
    path : str = "tidy3d.log"
        Download path to log file (including filename).
    verbose : bool = True
        If ``True``, will print progressbars and status, otherwise, will run silently.
    progress_callback : Callable[[float], None] = None
        Optional callback function called when downloading file with ``bytes_in_chunk`` as argument.

    Note
    ----
    To load downloaded results into data, call :meth:`load` with option ``replace_existing=False``.
    """
    task = SimulationTask(taskId=task_id)
    task.get_log(path, verbose=verbose, progress_callback=progress_callback)


@wait_for_connection
def load(
    task_id: TaskId,
    path: str = "simulation_data.hdf5",
    replace_existing: bool = True,
    verbose: bool = True,
    progress_callback: Optional[Callable[[float], None]] = None,
) -> WorkflowDataType:
    """
    Download and Load simulation results into :class:`.SimulationData` object.

    Notes
    -----

        After the simulation is complete, you can load the results into a :class:`.SimulationData` object by its
        ``task_id`` using:

        .. code-block:: python

            sim_data = web.load(task_id, path="outt/sim.hdf5", verbose=verbose)

        The :meth:`tidy3d.web.api.webapi.load` method is very convenient to load and postprocess results from simulations
        created using Tidy3D GUI.

    Parameters
    ----------
    task_id : str
        Unique identifier of task on server.  Returned by :meth:`upload`.
    path : str
        Download path to .hdf5 data file (including filename).
    replace_existing : bool = True
        Downloads the data even if path exists (overwriting the existing).
    verbose : bool = True
        If ``True``, will print progressbars and status, otherwise, will run silently.
    progress_callback : Callable[[float], None] = None
        Optional callback function called when downloading file with ``bytes_in_chunk`` as argument.

    Returns
    -------
    Union[:class:`.SimulationData`, :class:`.HeatSimulationData`, :class:`.EMESimulationData`]
        Object containing simulation data.
    """
    # For component modeler batches, default to a clearer filename if the default was used.
    if _is_modeler_batch(task_id) and os.path.basename(path) == "simulation_data.hdf5":
        base_dir = os.path.dirname(path) or "."
        path = os.path.join(base_dir, "cm_data.hdf5")

    if not os.path.exists(path) or replace_existing:
        download(task_id=task_id, path=path, verbose=verbose, progress_callback=progress_callback)

    if verbose:
        console = get_logging_console()
        if _is_modeler_batch(task_id):
            console.log(f"loading component modeler data from {path}")
        else:
            console.log(f"loading simulation from {path}")

    stub_data = Tidy3dStubData.postprocess(path)
    return stub_data


def _monitor_modeler_batch(
    batch_id: str,
    verbose: bool = True,
    max_detail_tasks: int = 20,
    worker_group: Optional[str] = None,
) -> None:
    """Monitor modeler batch progress with aggregate and per-task views."""
    console = get_logging_console() if verbose else None

    def _status_to_stage(status: str) -> tuple[str, int]:
        s = (status or "").lower()
        # Map a broader set of statuses to monotonic stages for progress bars
        if s in ("draft", "created"):
            return ("draft", 0)
        if s in ("queue", "queued"):
            return ("queued", 1)
        if s in ("preprocess",):
            return ("preprocess", 1)
        if s in ("validating",):
            return ("validating", 2)
        if s in ("validate_success", "validate_warn"):
            return ("validate", 3)
        if s in ("running",):
            return ("running", 4)
        if s in ("postprocess",):
            return ("postprocess", 5)
        if s in ("run_success", "success"):
            return ("success", 6)
        # Unknown statuses map to earliest stage to avoid showing 100% prematurely
        return (s or "unknown", 0)

    detail = _batch_detail(batch_id)
    name = detail.name or "modeler_batch"
    group_id = detail.groupId

    header = f"Subtasks status - {name}"
    if group_id:
        header += f"\nGroup ID: '{group_id}'"
    if console is not None:
        console.log(header)

    # Non-verbose path: poll without progress bars then return
    if not verbose:
        terminal_errors = {
            "validate_fail",
            "error",
            "diverged",
            "blocked",
            "aborting",
            "aborted",
        }
        # Run phase
        while True:
            d = _batch_detail(batch_id)
            s = d.totalStatus.value
            total = d.totalTask or 0
            r = d.runSuccess or 0
            if s in terminal_errors:
                raise WebError(f"Batch {batch_id} terminated: {s}")
            if total and r >= total:
                break
            time.sleep(REFRESH_TIME)
        # Postprocess phase
        BatchTask(batch_id).postprocess(batch_type="RF_SWEEP", worker_group=worker_group)
        while True:
            d = _batch_detail(batch_id)
            postprocess_status = d.postprocessStatus
            if postprocess_status == "success":
                break
            elif postprocess_status in terminal_errors:
                raise WebError(
                    f"Batch {batch_id} terminated. Please contact customer support and provide this Component Modeler batch ID: '{batch_id}'"
                )
            time.sleep(REFRESH_TIME)
        return

    progress_columns = (
        TextColumn("[progress.description]{task.description}"),
        BarColumn(bar_width=25),
        TaskProgressColumn(),
        TimeElapsedColumn(),
    )
    with Progress(*progress_columns, console=console, transient=False) as progress:
        terminal_errors = {"validate_fail", "error", "diverged", "blocked", "aborting", "aborted"}

        # Phase: Run (aggregate + per-task)
        p_run = progress.add_task("Run Total", total=1.0)
        task_bars: dict[str, int] = {}
        run_statuses = [
            "draft",
            "preprocess",
            "validating",
            "validate",
            "running",
            "postprocess",
            "success",
        ]

        while True:
            detail = _batch_detail(batch_id)
            status = detail.totalStatus.value
            total = detail.totalTask or 0
            r = detail.runSuccess or 0

            # Create per-task bars as soon as tasks appear
            if total and total <= max_detail_tasks and detail.tasks:
                name_to_task = {(t.taskName or t.taskId): t for t in (detail.tasks or [])}
                for name, t in name_to_task.items():
                    if name not in task_bars:
                        tstatus = (t.status or "draft").lower()
                        _, idx = _status_to_stage(tstatus)
                        pbar = progress.add_task(
                            f"  {name}",
                            total=len(run_statuses) - 1,
                            completed=min(idx, len(run_statuses) - 1),
                        )
                        task_bars[name] = pbar

            # Aggregate run progress: average stage fraction across tasks
            if detail.tasks:
                acc = 0.0
                n_members = 0
                for t in detail.tasks or []:
                    n_members += 1
                    tstatus = (t.status or "draft").lower()
                    _, idx = _status_to_stage(tstatus)
                    acc += max(0.0, min(1.0, idx / 6.0))
                run_frac = (acc / float(n_members)) if n_members else 0.0
            else:
                run_frac = (r / total) if total else 0.0
            progress.update(p_run, completed=run_frac)

            # Update per-task bars
            if task_bars and detail.tasks:
                name_to_task = {(t.taskName or t.taskId): t for t in (detail.tasks or [])}
                for tname, pbar in task_bars.items():
                    t = name_to_task.get(tname)
                    if not t:
                        continue
                    tstatus = (t.status or "draft").lower()
                    _, idx = _status_to_stage(tstatus)
                    completed = min(idx, 6)
                    desc = f"  {tname} [{tstatus or 'draft'}]"
                    progress.update(pbar, completed=completed, description=desc, refresh=False)

            if total and r >= total:
                break
            if status in terminal_errors:
                raise WebError(f"Batch {batch_id} terminated: {status}")
            progress.refresh()
            time.sleep(REFRESH_TIME)

        # Phase: Postprocess
        BatchTask(batch_id).postprocess(batch_type="RF_SWEEP", worker_group=worker_group)
        p_post = progress.add_task("Postprocess", total=1.0)
        while True:
            detail = _batch_detail(batch_id)
            postprocess_status = detail.postprocessStatus
            if postprocess_status == "success":
                progress.update(p_post, completed=1.0)
                progress.refresh()
                break
            elif postprocess_status == "queued":
                progress.update(p_post, completed=0.22)
            elif postprocess_status == "preprocess":
                progress.update(p_post, completed=0.33)
            elif postprocess_status == "running":
                progress.update(p_post, completed=0.55)
            elif postprocess_status in terminal_errors:
                raise WebError(
                    f"Batch {batch_id} terminated. Please contact customer support and provide this Component Modeler batch ID: '{batch_id}'"
                )
            progress.refresh()
            time.sleep(REFRESH_TIME)

        if console is not None:
            console.log("Modeler has finished running successfully.")
            real_cost(batch_id, verbose=verbose)


@wait_for_connection
def delete(task_id: TaskId, versions: bool = False) -> TaskInfo:
    """Delete server-side data associated with task.

    Parameters
    ----------
    task_id : str
        Unique identifier of task on server.  Returned by :meth:`upload`.
    versions : bool = False
        If ``True``, delete all versions of the task in the task group. Otherwise, delete only the version associated with the current task ID.

    Returns
    -------
    :class:`.TaskInfo`
        Object containing information about status, size, credits of task.

    """
    if not task_id:
        raise ValueError("Task id not found.")
    task = SimulationTask.get(task_id, verbose=False)
    task.delete(versions)
    return TaskInfo(**{"taskId": task.task_id, **task.dict()})


@wait_for_connection
def download_simulation(
    task_id: TaskId,
    path: str = SIM_FILE_HDF5,
    verbose: bool = True,
    progress_callback: Optional[Callable[[float], None]] = None,
) -> None:
    """Download the ``.hdf5`` file associated with the :class:`.Simulation` of a given task.

    Parameters
    ----------
    task_id : str
        Unique identifier of task on server.  Returned by :meth:`upload`.
    path : str = "simulation.hdf5"
        Download path to .hdf5 file of simulation (including filename).
    verbose : bool = True
        If ``True``, will print progressbars and status, otherwise, will run silently.
    progress_callback : Callable[[float], None] = None
        Optional callback function called when downloading file with ``bytes_in_chunk`` as argument.

    """
    task_info = get_info(task_id)
    task_type = task_info.taskType

    remote_sim_file = SIM_FILE_HDF5_GZ
    if task_type == "MODE_SOLVER":
        remote_sim_file = MODE_FILE_HDF5_GZ

    task = SimulationTask(taskId=task_id)
    task.get_simulation_hdf5(
        path, verbose=verbose, progress_callback=progress_callback, remote_sim_file=remote_sim_file
    )


@wait_for_connection
def get_tasks(
    num_tasks: Optional[int] = None, order: Literal["new", "old"] = "new", folder: str = "default"
) -> list[dict]:
    """Get a list with the metadata of the last ``num_tasks`` tasks.

    Parameters
    ----------
    num_tasks : int = None
        The number of tasks to return, or, if ``None``, return all.
    order : Literal["new", "old"] = "new"
        Return the tasks in order of newest-first or oldest-first.
    folder: str = "default"
        Folder from which to get the tasks.

    Returns
    -------
    List[Dict]
        List of dictionaries storing the information for each of the tasks last ``num_tasks`` tasks.
    """
    folder = Folder.get(folder, create=True)
    tasks = folder.list_tasks()
    if not tasks:
        return []
    if order == "new":
        tasks = sorted(tasks, key=lambda t: t.created_at, reverse=True)
    elif order == "old":
        tasks = sorted(tasks, key=lambda t: t.created_at)
    if num_tasks is not None:
        tasks = tasks[:num_tasks]
    return [task.dict() for task in tasks]


@wait_for_connection
def estimate_cost(
    task_id: str, verbose: bool = True, solver_version: Optional[str] = None
) -> float:
    """Compute the maximum FlexCredit charge for a given task.

    Parameters
    ----------
    task_id : str
        Unique identifier of task on server.  Returned by :meth:`upload`.
    verbose : bool = True
        Whether to log the cost and helpful messages.
    solver_version : str = None
        Target solver version.

    Returns
    -------
    float
        Estimated maximum cost for :class:`.Simulation` associated with given ``task_id``.

    Note
    ----
    Cost is calculated assuming the simulation runs for
    the full ``run_time``. If early shut-off is triggered, the cost is adjusted proportionately.
    A minimum simulation cost may also apply, which depends on the task details.

    Notes
    -----

        We can get the cost estimate of running the task before actually running it. This prevents us from
        accidentally running large jobs that we set up by mistake. The estimated cost is the maximum cost
        corresponding to running all the time steps.

    Examples
    --------

    Basic example:

    .. code-block:: python

        # initializes job, puts task on server (but doesn't run it)
        job = web.Job(simulation=sim, task_name="job", verbose=verbose)

        # estimate the maximum cost
        estimated_cost = web.estimate_cost(job.task_id)

        print(f'The estimated maximum cost is {estimated_cost:.3f} Flex Credits.')

    """
    if not isinstance(task_id, str):
        raise ValueError(
            f"Task ID: {task_id} is not a string. You can get it using 'web.upload(<WorkflowType>)'."
        )

    console = get_logging_console() if verbose else None

    if _is_modeler_batch(task_id):
        status = _batch_detail(task_id).totalStatus.value

        # Wait for a termination status
        while status not in ["validate_success", "success", "error", "failed"]:
            time.sleep(REFRESH_TIME)
            status = _batch_detail(task_id).totalStatus.value

        if status in ["validate_success", "success"]:
            est_flex_unit = _batch_detail(task_id).estFlexUnit
            if verbose:
                console.log(
                    f"Maximum FlexCredit cost: {est_flex_unit:1.3f}. Minimum cost depends on "
                    "task execution details. Use 'web.real_cost(task_id)' to get the billed FlexCredit "
                    "cost after a simulation run."
                )
            return est_flex_unit
    else:
        task = SimulationTask.get(task_id)

        if not task:
            raise ValueError("Task not found.")

        task.estimate_cost(solver_version=solver_version)
        task_info = get_info(task_id)
        status = task_info.metadataStatus

        # Wait for a termination status
        while status not in ["processed", "success", "error", "failed"]:
            time.sleep(REFRESH_TIME)
            task_info = get_info(task_id)
            status = task_info.metadataStatus

        if status in ["processed", "success"]:
            if verbose:
                console.log(
                    f"Estimated FlexCredit cost: {task_info.estFlexUnit:1.3f}. Minimum cost depends on "
                    "task execution details. Use 'web.real_cost(task_id)' to get the billed FlexCredit "
                    "cost after a simulation run."
                )
                fc_mode = task_info.estFlexCreditMode
                fc_post = task_info.estFlexCreditPostProcess
                if fc_mode:
                    console.log(f"  {fc_mode:1.3f} FlexCredit of the total cost from mode solves.")
                if fc_post:
                    console.log(
                        f"  {fc_post:1.3f} FlexCredit of the total cost from post-processing."
                    )
            return task_info.estFlexUnit

        # Something went wrong
        raise WebError("Could not get estimated cost!")


@wait_for_connection
def real_cost(task_id: str, verbose=True) -> float | None:
    """Get the billed cost for given task after it has been run.

    Parameters
    ----------
    task_id : str
        Unique identifier of task on server.  Returned by :meth:`upload`.
    verbose : bool = True
        Whether to log the cost and helpful messages.

    Returns
    -------
    float
        The flex credit cost that was billed for the given ``task_id``.

    Note
    ----
        The billed cost may not be immediately available when the task status is set to ``success``,
        but should be available shortly after.

    Examples
    --------

    To obtain the cost of a simulation, you can use the function ``tidy3d.web.real_cost(task_id)``. In the example
    below, a job is created, and its cost is estimated. After running the simulation, the real cost can be obtained.

    .. code-block:: python

        import time

        # initializes job, puts task on server (but doesn't run it)
        job = web.Job(simulation=sim, task_name="job", verbose=verbose)

        # estimate the maximum cost
        estimated_cost = web.estimate_cost(job.task_id)

        print(f'The estimated maximum cost is {estimated_cost:.3f} Flex Credits.')

        # Runs the simulation.
        sim_data = job.run(path="data/sim_data.hdf5")

        time.sleep(5)

        # Get the billed FlexCredit cost after a simulation run.
        cost = web.real_cost(job.task_id)
    """
    if not isinstance(task_id, str):
        raise ValueError(
            f"Task ID: {task_id} is not a string. You can get it using 'web.upload(<WorkflowType>)'."
        )

    console = get_logging_console() if verbose else None
    if _is_modeler_batch(task_id):
        status = _batch_detail(task_id).totalStatus.value
        flex_unit = _batch_detail(task_id).realFlexUnit or None
        if status not in ["success", "run_success"]:
            log.warning(
                f"Billed FlexCredit for task '{task_id}' is not available. If the task has been "
                "successfully run, it should be available shortly."
            )
        else:
            if verbose:
                console.log(
                    f"Billed FlexCredit cost: {flex_unit:1.3f}. Minimum cost depends on "
                    "task execution details. Use 'web.real_cost(task_id)' to get the billed FlexCredit "
                    "cost after a simulation run."
                )
        return flex_unit
    else:
        task_info = get_info(task_id)
        flex_unit = task_info.realFlexUnit
        ori_flex_unit = task_info.oriRealFlexUnit
        if not flex_unit:
            log.warning(
                f"Billed FlexCredit for task '{task_id}' is not available. If the task has been "
                "successfully run, it should be available shortly."
            )
        else:
            if verbose:
                console.log(f"Billed flex credit cost: {flex_unit:1.3f}.")
                if flex_unit != ori_flex_unit and task_info.taskType == "FDTD":
                    console.log(
                        "Note: the task cost pro-rated due to early shutoff was below the minimum "
                        "threshold, due to fast shutoff. Decreasing the simulation 'run_time' should "
                        "decrease the estimated, and correspondingly the billed cost of such tasks."
                    )
        return flex_unit


@wait_for_connection
def account(verbose=True) -> Account:
    """Get account information including FlexCredit balance and usage limits.

    Parameters
    ----------
    verbose : bool = True
        If ``True``, prints account information including credit balance, expiration,
        and free simulation counts.

    Returns
    -------
    Account
        Object containing account information such as credit balance, expiration dates,
        and daily free simulation counts.

    Examples
    --------
    Get account information:

    .. code-block:: python

        account_info = web.account()
        # Displays:
        # Current FlexCredit balance: 10.00 and expiration date: 2024-12-31 23:59:59.
        # Remaining daily free simulations: 3.
    """
    account_info = Account.get()
    if verbose and account_info:
        console = get_logging_console()
        credit = account_info.credit
        credit_expiration = account_info.credit_expiration
        cycle_type = account_info.allowance_cycle_type
        cycle_amount = account_info.allowance_current_cycle_amount
        cycle_end_date = account_info.allowance_current_cycle_end_date
        free_simulation_counts = account_info.daily_free_simulation_counts

        message = ""
        if credit is not None:
            message += f"Current FlexCredit balance: {credit:.2f}"
            if credit_expiration is not None:
                message += (
                    f" and expiration date: {credit_expiration.strftime('%Y-%m-%d %H:%M:%S')}. "
                )
            else:
                message += ". "
        if cycle_type is not None and cycle_amount is not None and cycle_end_date is not None:
            cycle_end = cycle_end_date.strftime("%Y-%m-%d %H:%M:%S")
            message += f"{cycle_type} FlexCredit balance: {cycle_amount:.2f} and expiration date: {cycle_end}. "
        if free_simulation_counts is not None:
            message += f"Remaining daily free simulations: {free_simulation_counts}."

        console.log(message)

    return account_info


@wait_for_connection
def test() -> None:
    """Confirm whether Tidy3D authentication is configured.

    Raises
    ------
    WebError
        If Tidy3D authentication is not configured correctly.

    Notes
    -----
    This method tests the authentication configuration by attempting to retrieve
    the task list. If authentication is not properly set up, it will raise an
    exception with instructions on how to configure authentication.

    Examples
    --------
    Test authentication:

    .. code-block:: python

        web.test()
        # If successful, displays:
        # Authentication configured successfully!
    """
    try:
        # note, this is a little slow, but the only call that doesn't require providing a task id.
        get_tasks(num_tasks=0)
        console = get_logging_console()
        console.log("Authentication configured successfully!")
    except (WebError, HTTPError) as e:
        url = "https://docs.flexcompute.com/projects/tidy3d/en/latest/index.html"
        msg = (
            str(e)
            + "\n\n"
            + "It looks like the Tidy3D Python interface is not configured with your "
            "unique API key. "
            "To get your API key, sign into 'https://tidy3d.simulation.cloud' and copy it "
            "from your 'Account' page. Then you can configure tidy3d through command line "
            "'tidy3d configure' (recommended). Alternatively, one can manually create the configuration "
            "file by creating a file at your home directory '~/.tidy3d/config' (unix) or "
            "'.tidy3d/config' (windows) with content like: \n\n"
            "apikey = 'XXX' \n\nHere XXX is your API key copied from your account page within quotes.\n\n"
            f"For details, check the instructions at {url}."
        )
        raise WebError(msg) from e



"""ZBF utilities"""

from __future__ import annotations

from struct import unpack

import numpy as np
import pydantic.v1 as pd

from tidy3d.components.base import Tidy3dBaseModel


class ZBFData(Tidy3dBaseModel):
    """
    Contains data read in from a ``.zbf`` file
    """

    version: int = pd.Field(title="Version", description="File format version number.")
    nx: int = pd.Field(title="Samples in X", description="Number of samples in the x direction.")
    ny: int = pd.Field(title="Samples in Y", description="Number of samples in the y direction.")
    ispol: bool = pd.Field(
        title="Is Polarized",
        description="``True`` if the beam is polarized, ``False`` otherwise.",
    )
    unit: str = pd.Field(
        title="Spatial Units", description="Spatial units, either 'mm', 'cm', 'in', or 'm'."
    )
    dx: float = pd.Field(title="Grid Spacing, X", description="Grid spacing in x.")
    dy: float = pd.Field(title="Grid Spacing, Y", description="Grid spacing in y.")
    zposition_x: float = pd.Field(
        title="Z Position, X Direction",
        description="The pilot beam z position with respect to the pilot beam waist, x direction.",
    )
    zposition_y: float = pd.Field(
        title="Z Position, Y Direction",
        description="The pilot beam z position with respect to the pilot beam waist, y direction.",
    )
    rayleigh_x: float = pd.Field(
        title="Rayleigh Distance, X Direction",
        description="The pilot beam Rayleigh distance in the x direction.",
    )
    rayleigh_y: float = pd.Field(
        title="Rayleigh Distance, Y Direction",
        description="The pilot beam Rayleigh distance in the y direction.",
    )
    waist_x: float = pd.Field(
        title="Beam Waist, X", description="The pilot beam waist in the x direction."
    )
    waist_y: float = pd.Field(
        title="Beam Waist, Y", description="The pilot beam waist in the y direction."
    )
    wavelength: float = pd.Field(..., title="Wavelength", description="The wavelength of the beam.")
    background_refractive_index: float = pd.Field(
        title="Background Refractive Index",
        description="The index of refraction in the current medium.",
    )
    receiver_eff: float = pd.Field(
        title="Receiver Efficiency",
        description="The receiver efficiency. Zero if fiber coupling is not computed.",
    )
    system_eff: float = pd.Field(
        title="System Efficiency",
        description="The system efficiency. Zero if fiber coupling is not computed.",
    )
    Ex: np.ndarray = pd.Field(
        title="Electric Field, X Component",
        description="Complex-valued electric field, x component.",
    )
    Ey: np.ndarray = pd.Field(
        title="Electric Field, Y Component",
        description="Complex-valued electric field, y component.",
    )

    def read_zbf(filename: str) -> ZBFData:
        """Reads a Zemax Beam File (``.zbf``)

        Parameters
        ----------
        filename : str
            The file name of the ``.zbf`` file to read.

        Returns
        -------
        :class:`.ZBFData`
        """

        # Read the zbf file
        with open(filename, "rb") as f:
            # Load the header
            version, nx, ny, ispol, units = unpack("<5I", f.read(20))
            f.read(16)  # unused values
            (
                dx,
                dy,
                zposition_x,
                rayleigh_x,
                waist_x,
                zposition_y,
                rayleigh_y,
                waist_y,
                wavelength,
                background_refractive_index,
                receiver_eff,
                system_eff,
            ) = unpack("<12d", f.read(96))
            f.read(64)  # unused values

            # read E field
            nsamps = 2 * nx * ny
            rawx = list(unpack(f"<{nsamps}d", f.read(8 * nsamps)))
            if ispol:
                rawy = list(unpack(f"<{nsamps}d", f.read(8 * nsamps)))

        # convert unit key to unit string
        map_units = {0: "mm", 1: "cm", 2: "in", 3: "m"}
        try:
            unit = map_units[units]
        except KeyError:
            raise KeyError(
                f"Invalid units specified in the zbf file (expected '0', '1', '2', or '3', got '{units}')."
            ) from None

        # load E field
        Ex_real = np.asarray(rawx[0::2]).reshape(nx, ny, order="F")
        Ex_imag = np.asarray(rawx[1::2]).reshape(nx, ny, order="F")
        if ispol:
            Ey_real = np.asarray(rawy[0::2]).reshape(nx, ny, order="F")
            Ey_imag = np.asarray(rawy[1::2]).reshape(nx, ny, order="F")
        else:
            Ey_real = np.zeros((nx, ny))
            Ey_imag = np.zeros((nx, ny))

        Ex = Ex_real + 1j * Ex_imag
        Ey = Ey_real + 1j * Ey_imag

        return ZBFData(
            version=version,
            nx=nx,
            ny=ny,
            ispol=ispol,
            unit=unit,
            dx=dx,
            dy=dy,
            zposition_x=zposition_x,
            zposition_y=zposition_y,
            rayleigh_x=rayleigh_x,
            rayleigh_y=rayleigh_y,
            waist_x=waist_x,
            waist_y=waist_y,
            wavelength=wavelength,
            background_refractive_index=background_refractive_index,
            receiver_eff=receiver_eff,
            system_eff=system_eff,
            Ex=Ex,
            Ey=Ey,
        )



"""Data structures for post-processing modal component simulations to calculate S-matrices."""

from __future__ import annotations

import pydantic.v1 as pd

from tidy3d.plugins.smatrix.component_modelers.modal import ModalComponentModeler
from tidy3d.plugins.smatrix.data.base import AbstractComponentModelerData
from tidy3d.plugins.smatrix.data.data_array import ModalPortDataArray


class ModalComponentModelerData(AbstractComponentModelerData):
    """A data container for the results of a :class:`.ModalComponentModeler` run.

    This class stores the original modeler and the simulation data obtained
    from running the simulations it defines. It also provides a method to
    compute the S-matrix from the simulation data.
    """

    modeler: ModalComponentModeler = pd.Field(
        ...,
        title="ModalComponentModeler",
        description="The original :class:`ModalComponentModeler` object that defines the simulation setup "
        "and from which this data was generated.",
    )

    def smatrix(self) -> ModalPortDataArray:
        """Computes and returns the scattering matrix (S-matrix).

        The S-matrix is computed from the simulation data using the
        :func:`.modal_construct_smatrix` function.

        Returns
        -------
        ModalPortDataArray
            The computed S-matrix.
        """
        from tidy3d.plugins.smatrix.analysis.modal import modal_construct_smatrix

        modal_port_data_array = modal_construct_smatrix(modeler_data=self)
        return modal_port_data_array



"""
This module defines the SimulationMap, a specialized container for managing
a set of Tidy3D simulations, allowing them to be accessed by name.
"""

from __future__ import annotations

from collections.abc import Iterator, Mapping
from typing import Any

import pydantic.v1 as pd

from tidy3d.components.base import Tidy3dBaseModel
from tidy3d.components.types.simulation import SimulationType


class ValueMap(Tidy3dBaseModel, Mapping[str, Any]):
    """An immutable dictionary-like container for objects.

    This class maps unique string keys to corresponding value objects.
    By inheriting from `collections.abc.Mapping`, it provides standard dictionary
    behaviors like item access (`my_dict["my_key"]`), iteration (`for name in my_dict`), and
    length checking (`len(my_dict)`).

    It automatically validates that the `keys` and `values`
    tuples have matching lengths upon instantiation.

    Attributes
    ----------
    keys : tuple[str, ...]
        A tuple of unique string identifiers for each simulation.
    values : tuple[Any, ...]
        A tuple of `Any`-type objects, each corresponding to a key at the
        same index. Should be overwritten by the subclass instantiation
    """

    keys_tuple: tuple[str, ...] = pd.Field(
        description="A tuple of unique string identifiers for each simulation.", alias="keys"
    )
    values_tuple: tuple[Any, ...] = pd.Field(
        description=(
            "A tuple of `Simulation` objects, each corresponding to a key at the same index."
        ),
        alias="values",
    )

    @pd.root_validator(skip_on_failure=True)
    def _validate_lengths_match(cls, data: dict) -> dict:
        """Pydantic root validator to ensure 'keys' and 'values' have the same length.

        Parameters
        ----------
        data : dict
            The dictionary of field values for the model provided by Pydantic.

        Returns
        -------
        dict
            The validated dictionary of field values.

        Raises
        ------
        ValueError
            If the lengths of the 'keys' and 'values' tuples are not equal.
        """
        keys, values = data.get("keys"), data.get("values")
        if keys is not None and values is not None and len(keys) != len(values):
            raise ValueError("Length of 'keys' and 'values' must be the same.")
        return data

    def __getitem__(self, key: str) -> Any:
        """Retrieves a `Simulation` object by its corresponding key.

        This allows for dictionary-style access, e.g., `my_map["my_key"]`.

        Parameters
        ----------
        key : str
            The string name of the item to retrieve.

        Returns
        -------
        SimulationType
            The object corresponding to the given key.

        Raises
        ------
        KeyError
            If no object with the given key name is found in the map.
        """
        for i, current_key in enumerate(self.keys_tuple):
            if current_key == key:
                return self.values_tuple[i]
        raise KeyError(f"Key '{key}' not found in the {type(self)}.")

    def __iter__(self) -> Iterator[str]:
        """Returns an iterator over the string keys of the map.

        This allows for standard iteration, e.g., `for key in my_map:`.

        Yields
        ------
        str
            The next key in the map.
        """
        return iter(self.keys_tuple)

    def __len__(self) -> int:
        """Returns the number of key-value pairs stored in the map.

        This allows for using the built-in `len()` function, e.g., `len(my_map)`.

        Returns
        -------
        int
            The total number of items in the map.
        """
        return len(self.keys_tuple)


class SimulationMap(ValueMap, Mapping[str, SimulationType]):
    """An immutable dictionary-like container for simulations.

    This class maps unique string keys to corresponding `Simulation` objects.
    By inheriting from `collections.abc.Mapping`, it provides standard dictionary
    behaviors like item access (`sims["sim_A"]`), iteration (`for name in sims`), and
    length checking (`len(sims)`).

    It automatically validates that the `keys` and `values`
    tuples have matching lengths upon instantiation.

    Attributes
    ----------
    keys : tuple[str, ...]
        A tuple of unique string identifiers for each simulation.
    values : tuple[SimulationType, ...]
        A tuple of `Simulation` objects, each corresponding to a key at the
        same index.

    Example
    -------
    >>> from tidy3d import (
    ...     Simulation,
    ...     SimulationMap,
    ...     Structure,
    ...     Box,
    ...     Medium,
    ...     UniformCurrentSource,
    ...     GaussianPulse,
    ...     FieldMonitor,
    ...     GridSpec,
    ...     BoundarySpec,
    ...     PML,
    ... )
    >>> # Simple minimal simulation
    >>> sim1 = Simulation(
    ...     size=(4, 3, 3),
    ...     grid_spec=GridSpec.auto(min_steps_per_wvl=25),
    ...     structures=[
    ...         Structure(
    ...             geometry=Box(size=(1, 1, 1), center=(0, 0, 0)),
    ...             medium=Medium(permittivity=2.0),
    ...         ),
    ...     ],
    ...     sources=[
    ...         UniformCurrentSource(
    ...             size=(0, 0, 0),
    ...             center=(0, 0.5, 0),
    ...             polarization="Hx",
    ...             source_time=GaussianPulse(freq0=2e14, fwidth=4e13),
    ...         )
    ...     ],
    ...     monitors=[
    ...         FieldMonitor(
    ...             size=(1, 1, 0),
    ...             center=(0, 0, 0),
    ...             freqs=[2e14],
    ...             name='field'
    ...         ),
    ...     ],
    ...     run_time=1e-12,
    ...     boundary_spec=BoundarySpec.all_sides(boundary=PML()),
    ... )
    >>> sim2 = sim1.updated_copy(run_time=2e-12)
    >>>
    >>> # Instantiate the map
    >>> simulation_map = SimulationMap(
    ...     keys=("sim_1", "sim_2"),
    ...     values=(sim1, sim2),
    ... )
    >>>
    >>> # Access a simulation like a dictionary
    >>> # print(simulation_map["sim_1"])
    """

    keys_tuple: tuple[str, ...] = pd.Field(
        description="A tuple of unique string identifiers for each simulation.", alias="keys"
    )
    values_tuple: tuple[SimulationType, ...] = pd.Field(
        description=(
            "A tuple of `Simulation` objects, each corresponding to a key at the same index."
        ),
        alias="values",
    )



"""Defines specification for apodization."""

from __future__ import annotations

import numpy as np
import pydantic.v1 as pd

from tidy3d.constants import SECOND
from tidy3d.exceptions import SetupError

from .base import Tidy3dBaseModel, skip_if_fields_missing
from .types import ArrayFloat1D, Ax
from .viz import add_ax_if_none


class ApodizationSpec(Tidy3dBaseModel):
    """Stores specifications for the apodizaton of frequency-domain monitors.

    Example
    -------
    >>> apod_spec = ApodizationSpec(start=1, end=2, width=0.2)


    .. image:: ../../_static/img/apodization.png
        :width: 80%
        :align: center

    """

    start: pd.NonNegativeFloat = pd.Field(
        None,
        title="Start Interval",
        description="Defines the time at which the start apodization ends.",
        units=SECOND,
    )

    end: pd.NonNegativeFloat = pd.Field(
        None,
        title="End Interval",
        description="Defines the time at which the end apodization begins.",
        units=SECOND,
    )

    width: pd.PositiveFloat = pd.Field(
        None,
        title="Apodization Width",
        description="Characteristic decay length of the apodization function, i.e., the width of the ramping up of the scaling function from 0 to 1.",
        units=SECOND,
    )

    @pd.validator("end", always=True, allow_reuse=True)
    @skip_if_fields_missing(["start"])
    def end_greater_than_start(cls, val, values):
        """Ensure end is greater than or equal to start."""
        start = values.get("start")
        if val is not None and start is not None and val < start:
            raise SetupError("End apodization begins before start apodization ends.")
        return val

    @pd.validator("width", always=True, allow_reuse=True)
    @skip_if_fields_missing(["start", "end"])
    def width_provided(cls, val, values):
        """Check that width is provided if either start or end apodization is requested."""
        start = values.get("start")
        end = values.get("end")
        if (start is not None or end is not None) and val is None:
            raise SetupError("Apodization width must be set.")
        return val

    @add_ax_if_none
    def plot(self, times: ArrayFloat1D, ax: Ax = None) -> Ax:
        """Plot the apodization function.

        Parameters
        ----------
        times : np.ndarray
            Array of times (seconds) to plot source at.
            To see source time amplitude for a specific :class:`.Simulation`,
            pass ``simulation.tmesh``.
        ax : matplotlib.axes._subplots.Axes = None
            Matplotlib axes to plot on, if not specified, one is created.

        Returns
        -------
        matplotlib.axes._subplots.Axes
            The supplied or created matplotlib axes.
        """
        times = np.array(times)
        amp = np.ones_like(times)

        if self.start is not None:
            start_ind = times < self.start
            time_scaled = (times[start_ind] - self.start) / self.width
            amp[start_ind] *= np.exp(-0.5 * time_scaled**2)

        if self.end is not None:
            end_ind = times > self.end
            time_scaled = (times[end_ind] - self.end) / self.width
            amp[end_ind] *= np.exp(-0.5 * time_scaled**2)

        ax.plot(times, amp, color="blueviolet")
        ax.set_xlabel("time (s)")
        ax.set_title("apodization function")
        ax.set_aspect("auto")
        return ax



from __future__ import annotations

from typing import Any

import pydantic.v1 as pd

from .base import Expression
from .types import NumberOrExpression, NumberType


class UnaryOperator(Expression):
    """
    Base class for unary operators in the metrics module.

    This class represents an operation with a single operand.
    Subclasses should implement the evaluate method to define the specific operation.
    """

    operand: NumberOrExpression = pd.Field(
        ...,
        title="Operand",
        description="The operand for the unary operator.",
    )

    _symbol: str
    _format: str = "({symbol}{operand})"

    @pd.validator("operand", pre=True, always=True)
    def validate_operand(cls, v):
        return cls._to_expression(v)

    def __repr__(self) -> str:
        return self._format.format(symbol=self._symbol, operand=self.operand)


class BinaryOperator(Expression):
    """
    Base class for binary operators in the metrics module.

    This class represents an operation with two operands.
    Subclasses should implement the evaluate method to define the specific operation.
    """

    left: NumberOrExpression = pd.Field(
        ...,
        title="Left",
        description="The left operand for the binary operator.",
    )
    right: NumberOrExpression = pd.Field(
        ...,
        title="Right",
        description="The right operand for the binary operator.",
    )

    _symbol: str
    _format: str = "({left} {symbol} {right})"

    @pd.validator("left", "right", pre=True, always=True)
    def validate_operands(cls, v):
        return cls._to_expression(v)

    def __repr__(self) -> str:
        return self._format.format(left=self.left, symbol=self._symbol, right=self.right)


class Add(BinaryOperator):
    """
    Represents the addition operation.
    """

    _symbol: str = "+"

    def evaluate(self, *args: Any, **kwargs: Any) -> NumberType:
        return self.left(*args, **kwargs) + self.right(*args, **kwargs)


class Subtract(BinaryOperator):
    """
    Represents the subtraction operation.
    """

    _symbol: str = "-"

    def evaluate(self, *args: Any, **kwargs: Any) -> NumberType:
        return self.left(*args, **kwargs) - self.right(*args, **kwargs)


class Multiply(BinaryOperator):
    """
    Represents the multiplication operation.
    """

    _symbol: str = "*"

    def evaluate(self, *args: Any, **kwargs: Any) -> NumberType:
        return self.left(*args, **kwargs) * self.right(*args, **kwargs)


class Negate(UnaryOperator):
    """
    Represents the negation operation.
    """

    _symbol: str = "-"

    def evaluate(self, *args: Any, **kwargs: Any) -> NumberType:
        return -self.operand(*args, **kwargs)


class Abs(UnaryOperator):
    """
    Represents the absolute value operation.
    """

    _symbol: str = "abs"
    _format = "{symbol}({operand})"

    def evaluate(self, *args: Any, **kwargs: Any) -> NumberType:
        return abs(self.operand(*args, **kwargs))


class Divide(BinaryOperator):
    """
    Represents the division operation.
    """

    _symbol: str = "/"

    def evaluate(self, *args: Any, **kwargs: Any) -> NumberType:
        return self.left(*args, **kwargs) / self.right(*args, **kwargs)


class Power(BinaryOperator):
    """
    Represents the exponentiation operation.
    """

    _symbol: str = "**"

    def evaluate(self, *args: Any, **kwargs: Any) -> NumberType:
        return self.left(*args, **kwargs) ** self.right(*args, **kwargs)


class Modulus(BinaryOperator):
    """
    Represents the modulus operation.
    """

    _symbol: str = "%"

    def evaluate(self, *args: Any, **kwargs: Any) -> NumberType:
        return self.left(*args, **kwargs) % self.right(*args, **kwargs)


class FloorDivide(BinaryOperator):
    """
    Represents the floor division operation.
    """

    _symbol: str = "//"

    def evaluate(self, *args: Any, **kwargs: Any) -> NumberType:
        return self.left(*args, **kwargs) // self.right(*args, **kwargs)


class MatMul(BinaryOperator):
    """
    Represents the matrix multiplication operation.
    """

    _symbol: str = "@"

    def evaluate(self, *args: Any, **kwargs: Any) -> NumberType:
        return self.left(*args, **kwargs) @ self.right(*args, **kwargs)



from __future__ import annotations

from typing import Any

import pydantic.v1 as pd
from numpy import inf

from tidy3d.components.base import Tidy3dBaseModel


class AbstractPlotParams(Tidy3dBaseModel):
    """Abstract class for storing plotting parameters.
    Corresponds with select properties of ``matplotlib.artist.Artist``.
    """

    alpha: Any = pd.Field(1.0, title="Opacity")
    zorder: float = pd.Field(None, title="Display Order")

    def include_kwargs(self, **kwargs) -> AbstractPlotParams:
        """Update the plot params with supplied kwargs."""
        update_dict = {
            key: value
            for key, value in kwargs.items()
            if key not in ("type",) and value is not None and key in self.__fields__
        }
        return self.copy(update=update_dict)

    def override_with_viz_spec(self, viz_spec) -> AbstractPlotParams:
        """Override plot params with supplied VisualizationSpec."""
        return self.include_kwargs(**dict(viz_spec))

    def to_kwargs(self) -> dict:
        """Export the plot parameters as kwargs dict that can be supplied to plot function."""
        kwarg_dict = self.dict()
        for ignore_key in ("type", "attrs"):
            kwarg_dict.pop(ignore_key)
        return kwarg_dict


class PathPlotParams(AbstractPlotParams):
    """Stores plotting parameters / specifications for a path.
    Corresponds with select properties of ``matplotlib.lines.Line2D``.
    """

    color: Any = pd.Field(None, title="Color", alias="c")
    linewidth: pd.NonNegativeFloat = pd.Field(2, title="Line Width", alias="lw")
    linestyle: str = pd.Field("--", title="Line Style", alias="ls")
    marker: Any = pd.Field("o", title="Marker Style")
    markeredgecolor: Any = pd.Field(None, title="Marker Edge Color", alias="mec")
    markerfacecolor: Any = pd.Field(None, title="Marker Face Color", alias="mfc")
    markersize: pd.NonNegativeFloat = pd.Field(10, title="Marker Size", alias="ms")


class PlotParams(AbstractPlotParams):
    """Stores plotting parameters / specifications for a given model.
    Corresponds with select properties of ``matplotlib.patches.Patch``.
    """

    edgecolor: Any = pd.Field(None, title="Edge Color", alias="ec")
    facecolor: Any = pd.Field(None, title="Face Color", alias="fc")
    fill: bool = pd.Field(True, title="Is Filled")
    hatch: str = pd.Field(None, title="Hatch Style")
    linewidth: pd.NonNegativeFloat = pd.Field(1, title="Line Width", alias="lw")


# defaults for different tidy3d objects
plot_params_geometry = PlotParams()
plot_params_structure = PlotParams()
plot_params_source = PlotParams(alpha=0.4, facecolor="limegreen", edgecolor="limegreen", lw=3)
plot_params_absorber = PlotParams(
    alpha=0.4, facecolor="lightskyblue", edgecolor="lightskyblue", lw=3
)
plot_params_monitor = PlotParams(alpha=0.4, facecolor="orange", edgecolor="orange", lw=3)
plot_params_pml = PlotParams(alpha=0.7, facecolor="gray", edgecolor="gray", hatch="x", zorder=inf)
plot_params_pec = PlotParams(alpha=1.0, facecolor="gold", edgecolor="black", zorder=inf)
plot_params_pmc = PlotParams(alpha=1.0, facecolor="lightsteelblue", edgecolor="black", zorder=inf)
plot_params_bloch = PlotParams(alpha=1.0, facecolor="orchid", edgecolor="black", zorder=inf)
plot_params_abc = PlotParams(alpha=1.0, facecolor="lightskyblue", edgecolor="black", zorder=inf)
plot_params_symmetry = PlotParams(edgecolor="gray", facecolor="gray", alpha=0.6, zorder=inf)
plot_params_override_structures = PlotParams(
    linewidth=0.4, edgecolor="black", fill=False, zorder=inf
)
plot_params_fluid = PlotParams(facecolor="white", edgecolor="lightsteelblue", lw=0.4, hatch="xx")
plot_params_grid = PlotParams(edgecolor="black", lw=0.2)
plot_params_lumped_element = PlotParams(
    alpha=0.4, facecolor="mediumblue", edgecolor="mediumblue", lw=3
)
